import sys
import os
import meerkathi.dispatch_crew.utils as utils
import yaml
import stimela.dismissable as sdm

NAME = "Cross calibration"

# Rules for interpolation mode to use when applying calibration solutions
applycal_interp_rules = {
   'bpcal'    :  {
                  'delay_cal'      : 'nearest', 
                  'bp_cal'         : 'nearest', 
                  'transfer_fluxscale': 'linear',
                  'gain_cal_gain'  : 'linear',
                 },
   'gcal'     :  {
                  'delay_cal'      : 'linear', 
                  'bp_cal'         : 'linear', 
                  'transfer_fluxscale': 'nearest',
                  'gain_cal_gain'  : 'linear',
                 },
   'target'   :  {
                  'delay_cal'      : 'linear', 
                  'bp_cal'         : 'linear', 
                  'transfer_fluxscale': 'linear',
                  'gain_cal_gain'  : 'linear',
                 },
}

table_suffix = {
    "delay_cal"             : 'K0',
    "bp_cal"                : 'B0', 
    "gain_cal_gain"         : 'G0', 
    "gain_cal_flux"         : 'G0', 
    "transfer_fluxscale"    : 'F0', 
}


def worker(pipeline, recipe, config):

    for i in range(pipeline.nobs):

        msname = pipeline.msnames[i]
        refant = pipeline.reference_antenna[i] or '0'
        prefix = pipeline.prefixes[i]
        dataid = pipeline.dataid[i]
        msinfo = '{0:s}/{1:s}-obsinfo.json'.format(pipeline.output, prefix)
        prefix = '{0:s}-{1:s}'.format(prefix, config.get('label', ''))

        # Check if field was specified as known key, else return the 
        # same value. 
        def get_field(field):
            """
                gets field ids parsed previously in the pipeline 
                params:
                    field: list of ids or comma-seperated list of ids where
                           ids are in bpcal, gcal, target, fcal or an actual field name
            """
            return ','.join(map(lambda x: ','.join(getattr(pipeline, x)[i].split(',')
                                                if isinstance(getattr(pipeline, x)[i], str) else getattr(pipeline, x)[i])
                                              if x in ['bpcal', 'gcal', 'target', 'fcal']
                                              else x.split(','),
                                field.split(',') if isinstance(field, str) else field))

        def get_gain_field(applyme, applyto=None):
            if applyme == 'delay_cal':
                return get_field(config['delay_cal'].get('field', 'bpcal,gcal'))
            if applyme == 'bp_cal':
                return get_field('bpcal')
            if applyme == 'gain_cal_flux':
                return get_field('fcal')
            if applyme == 'gain_cal_gain':
                return get_field('gcal')
            if applyme == 'transfer_fluxscale':
                if applyto in ['gcal', 'target']:
                    return get_field('gcal')
                elif applyto == 'bpcal':
                    return get_field('bpcal')

        def flag_gains(cal, opts, datacolumn="CPARAM"):
            opts = dict(opts)
            step = 'plot_{0:s}_{1:d}'.format(cal, i)
            opts["vis"] = '{0:s}.{1:s}:output'.format(prefix, table_suffix[cal])
            opts["datacolumn"] = datacolumn
            recipe.add('cab/casa_flagdata', step, opts,
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Flagging gains'.format(step))

        if pipeline.enable_task(config, 'set_model'):
            # Set model
            field = get_field(config['set_model'].get('field', 'fcal'))
            if config['set_model'].get('no_verify', False):
                opts = {
                    "vis"        : msname,
                    "field"      : field,
                    "scalebychan": True,
                    "usescratch" : True,
                }
            else:
                model = utils.find_in_native_calibrators(msinfo, field)
                standard = utils.find_in_casa_calibrators(msinfo, field)
                # Prefer our standard over the NRAO standard
                if model:
                    opts = {
                      "vis"         : msname,
                      "field"       : field,
                      "standard"    : "manual",
                      "fluxdensity" : model['I'],
                      "reffreq"     : '{0:f}GHz'.format(model['ref']/1e9),
                      "spix"        : [model[a] for a in 'abcd'],
                      "scalebychan" : True,
                      "usescratch"  : False,
                    }
                elif standard:
                   opts = {
                      "vis"         : msname,
                      "field"       : field,
                      "standard"    : config['set_model'].get('standard', standard),
                      "usescratch"  : False,
                      "scalebychan" : True,
                    }
                else:
                    raise RuntimeError('The flux calibrator field "{}" could not be \
found in our database or in the CASA NRAO database'.format(field))
            step = 'set_model_cal_{0:d}'.format(i)
            recipe.add('cab/casa_setjy', step,
               opts,
               input=pipeline.input,
               output=pipeline.output,
               label='{0:s}:: Set jansky ms={1:s}'.format(step, msname))
        # Delay calibration
        if pipeline.enable_task(config, 'delay_cal'):
            step = 'delay_cal_{0:d}'.format(i)
            field = get_field(config['delay_cal'].get('field', 'bpcal,gcal'))
            recipe.add('cab/casa_gaincal', step,
               {
                 "vis"          : msname,
                 "caltable"     : prefix+".K0",
                 "field"        : field,
                 "refant"       : refant,
                 "solint"       : config['delay_cal'].get('solint', 'inf'),
                 "combine"      : config['delay_cal'].get('combine', ''),
                 "gaintype"     : "K",
                 "uvrange"      : config.get('uvrange', ''),
               },
               input=pipeline.input,
               output=pipeline.output,
               label='{0:s}:: Delay calibration ms={1:s}'.format(step, msname))

            if config['delay_cal'].get('flag', {"enabled": False}).get('enabled', False):
                flag_gains('delay_cal', config['delay_cal']['flag'], datacolumn="FPARAM")

            if  config['delay_cal'].get('plot', True):
                step = 'plot_delay_cal_{0:d}'.format(i)
                recipe.add('cab/casa_plotcal', step,
                   {
                    "caltable"  : prefix+".K0:output",
                    "xaxis"     : 'antenna',
                    "yaxis"     : 'delay',
                    "field"     : field,
                    "iteration" : 'antenna',
                    "subplot"   : 441,
                    "plotsymbol": 'o',
                    "figfile"   : '{0:s}-K0.png'.format(prefix),
                    "showgui"   : False,
                   },
                   input=pipeline.input,
                   output=pipeline.output,
                   label='{0:s}:: Plot gaincal phase ms={1:s}'.format(step, msname))

        # Set "Combine" to 'scan' for getting combining all scans for BP soln.
        if pipeline.enable_task(config, 'bp_cal'):
            if config.get('otfdelay', True): gaintables,interpolations=[prefix+'.K0:output'],['nearest']
            else: gaintables,interpolations=None,''
            field = get_field(config['bp_cal'].get('field', 'bpcal'))
            step = 'bp_cal_{0:d}'.format(i)

            recipe.add('cab/casa_bandpass', step,
               {
                 "vis"          : msname,
                 "caltable"     : prefix+'.B0',
                 "field"        : field,
                 "refant"       : refant,
                 "solint"       : config['bp_cal'].get('solint', 'inf'),
                 "combine"      : config['bp_cal'].get('combine', ''),
                 "bandtype"     : "B",
                 "gaintable"    : sdm.dismissable(gaintables),
                 "interp"       : interpolations,
                 "fillgaps"     : 70,
                 "uvrange"      : config['uvrange'],
                 "minsnr"       : config['bp_cal'].get('minsnr', 5),
                 "minblperant"  : config['bp_cal'].get('minnrbl', 4),
                 "solnorm"      : config['bp_cal'].get('solnorm', False),
               },
               input=pipeline.input,
               output=pipeline.output,
               label='{0:s}:: Bandpass calibration ms={1:s}'.format(step, msname))

            if config['bp_cal'].get('flag', {"enabled": False}).get('enabled', False):
                flag_gains('bp_cal', config['bp_cal']['flag'])

            if config['bp_cal'].get('plot', True):
                 for plot in 'amp','phase':
                    step = 'plot_bandpass_{0:s}_{1:d}'.format(plot, i)
                    recipe.add('cab/casa_plotcal', step,
                       {
                        "caltable"  : prefix+".B0:output",
                        "xaxis"     : 'freq',
                        "yaxis"     : plot,
                        "field"     : field,
                        "iteration" : 'antenna',
                        "subplot"   : 441,
                        "plotsymbol": ',',
                        "figfile"   : '{0:s}-B0-{1:s}.png'.format(prefix, plot),
                        "showgui"   : False,
                       },
                       input=pipeline.input,
                       output=pipeline.output,
                       label='{0:s}:: plot bandpass calibration gain caltable={1:s}'.format(step, prefix+".B0:output"))

        # Gain calibration for Flux calibrator field
        if pipeline.enable_task(config, 'gain_cal_flux'):
            if config.get('otfdelay', True): gaintables,interpolations=[prefix+'.K0:output'],['nearest']
            else: gaintables,interpolations=[],[]
            gaintables+=[prefix+".B0:output"]
            interpolations+=['nearest']
            step = 'gain_cal_flux_{0:d}'.format(i)
            field = get_field(config['gain_cal_flux'].get('field', 'fcal'))
            recipe.add('cab/casa_gaincal', step,
               {
                 "vis"          : msname,
                 "caltable"     : prefix+".G0:output",
                 "field"        : field,
                 "refant"       : refant,
                 "solint"       : config['gain_cal_flux'].get('solint', 'inf'),
                 "combine"      : config['gain_cal_flux'].get('combine', ''),
                 "gaintype"     : "G",
                 "calmode"      : 'ap',
                 "gaintable"    : gaintables,
                 "interp"       : interpolations,
                 "uvrange"      : config['uvrange'],
                 "minsnr"       : config['gain_cal_flux'].get('minsnr', 5),
                 "minblperant"  : config['gain_cal_flux'].get('minnrbl', 4),
               },
               input=pipeline.input,
               output=pipeline.output,
               label='{0:s}:: Gain calibration for bandpass ms={1:s}'.format(step, msname))

            if config['gain_cal_flux'].get('plot', True):
                for plot in 'amp','phase':
                    step = 'plot_gain_cal_flux_{0:s}_{1:d}'.format(plot, i)
                    recipe.add('cab/casa_plotcal', step,
                       {
                        "caltable"  : prefix+".G0:output",
                        "xaxis"     : 'time',
                        "yaxis"     : plot,
                        "field"     : field,
                        "iteration" : 'antenna',
                        "subplot"   : 441,
                        "plotsymbol": 'o',
                        "figfile"   : '{0:s}-G0-fcal-{1:s}.png'.format(prefix, plot),
                        "showgui"   : False,
                       },
                       input=pipeline.input,
                       output=pipeline.output,
                       label='{0:s}:: Plot gaincal phase ms={1:s}'.format(step, msname))

            if config['gain_cal_flux'].get('flag', {"enable": False}).get('enable', False):
                flag_gains('gain_cal_flux', config['gain_cal_flux']['flag'])

        # Gain calibration for Gaincal field
        if pipeline.enable_task(config, 'gain_cal_gain'):
            if config.get('otfdelay', True): gaintables,interpolations=[prefix+'.K0:output'],['linear']
            else: gaintables,interpolations=[],[]
            gaintables+=[prefix+".B0:output"]
            interpolations+=['linear']
            step = 'gain_cal_gain_{0:d}'.format(i)
            field = get_field(config['gain_cal_gain'].get('field', 'gcal'))
            recipe.add('cab/casa_gaincal', step,
               {
                 "vis"          : msname,
                 "caltable"     : prefix+".G0:output",
                 "field"        : field,
                 "refant"       : refant,
                 "solint"       : config['gain_cal_gain'].get('solint', 'inf'),
                 "combine"      : config['gain_cal_gain'].get('combine', ''),
                 "gaintype"     : "G",
                 "calmode"      : 'ap',
                 "gaintable"    : gaintables,
                 "interp"       : interpolations,
                 "uvrange"      : config['uvrange'],
                 "minsnr"       : config['gain_cal_gain'].get('minsnr', 5),
                 "minblperant"  : config['gain_cal_gain'].get('minnrbl', 4),
                 "append"       : True,
               },
               input=pipeline.input,
               output=pipeline.output,
               label='{0:s}:: Gain calibration ms={1:s}'.format(step, msname))

            if config['gain_cal_gain'].get('flag', {"enabled": False}).get('enabled', False):
                flag_gains('gain_cal_gain', config['gain_cal_gain']['flag'])

            if config['gain_cal_gain'].get('plot', True):
                for plot in 'amp','phase':
                    step = 'plot_gain_cal_{0:s}_{1:d}'.format(plot, i)
                    recipe.add('cab/casa_plotcal', step,
                       {
                        "caltable"  : prefix+".G0:output",
                        "xaxis"     : 'time',
                        "yaxis"     : plot,
                        "field"     : field,
                        "iteration" : 'antenna',
                        "subplot"   : 441,
                        "plotsymbol": 'o',
                        "figfile"   : '{0:s}-G0-gcal-{1:s}.png'.format(prefix, plot),
                        "showgui"   : False,
                       },
                       input=pipeline.input,
                       output=pipeline.output,
                       label='{0:s}:: Plot gaincal phase ms={1:s}'.format(step, msname))

        # Flux scale transfer
        if pipeline.enable_task(config, 'transfer_fluxscale'):
            ref = get_field(config['transfer_fluxscale'].get('reference', 'fcal'))
            trans = get_field(config['transfer_fluxscale'].get('transfer', 'gcal'))
            step = 'transfer_fluxscale_{0:d}'.format(i)
            recipe.add('cab/casa_fluxscale', step,
               {
                "vis"          : msname,
                "caltable"      : prefix+".G0:output",
                "fluxtable"     : prefix+".F0:output",
                "reference"     : [ref],
                "transfer"      : [trans],
               },
               input=pipeline.input,
               output=pipeline.output,
               label='{0:s}:: Flux scale transfer ms={1:s}'.format(step, msname))

            if config['transfer_fluxscale'].get('plot', True):
                for plot in 'amp','phase':
                    step = 'plot_fluxscale_{0:s}_{1:d}'.format(plot, i)
                    recipe.add('cab/casa_plotcal', step,
                       {
                        "caltable"  : prefix+".F0:output",
                        "xaxis"     : 'time',
                        "yaxis"     : plot,
                        "field"     : '',
                        "iteration" : 'antenna',
                        "subplot"   : 441,
                        "plotsymbol": 'o',
                        "figfile"   : '{0:s}-F0-{1:s}.png'.format(prefix, plot),
                        "showgui"   : False,
                       },
                       input=pipeline.input,
                       output=pipeline.output,
                       label='{0:s}:: Plot gaincal phase ms={1:s}'.format(step, msname))

        applied = []
        for ft in ['bpcal','gcal','target']:
            gaintablelist,gainfieldlist,interplist = [],[],[]
            no_table_to_apply = True
            field = getattr(pipeline, ft)[i]
            for applyme in 'delay_cal bp_cal gain_cal_flux gain_cal_gain transfer_fluxscale'.split():
                if not pipeline.enable_task(config, 'apply_'+applyme):
                    continue
                if ft not in config['apply_'+applyme].get('applyto', 'bpcal,gcal,target').split(','):
                   continue
                suffix = table_suffix[applyme]
                interp = applycal_interp_rules[ft][applyme]
                gainfield = get_gain_field(applyme, ft)
                gaintablelist.append(prefix+'.{:s}:output'.format(suffix))
                gainfieldlist.append(gainfield)
                interplist.append(interp)
                no_table_to_apply = False

            if no_table_to_apply or field in applied:
                continue

            applied.append(field)
            step = 'apply_{0:s}_{1:d}'.format(ft, i)
            recipe.add('cab/casa_applycal', step,
               {
                "vis"       : msname,
                "field"     : field,
                "gaintable" : gaintablelist,
                "gainfield" : gainfieldlist,
                "interp"    : interplist,
                "calwt"     : [False],
                "parang"    : False,
                "applymode" : config['apply_'+applyme]['applymode'],
               },
               input=pipeline.input,
               output=pipeline.output,
               label='{0:s}:: Apply calibration to field={1:s}, ms={2:s}'.format(step, field, msname))

        # auto flag closure errors and systematic issues based on calibrated calibrator phase (chi-squared thresholds)
        # Physical assertion: Expect calibrator phase == 0 (calibrator at phase centre). Corrected phases should be at phase centre
        # Any deviation from this expected value per baseline means that there are baseline-based gains (closure errors)
        # that couldn't be corrected solely antenna-based gains. These baselines must be flagged out of the calibrator and the 
        # target fields between calibrator scans which were marked bad.
        # Compare in-between scans per baseline per field per channel
        # Also compare in-between baselines per scan per field per channel
        if pipeline.enable_task(config, 'autoflag_closure_error'):
            step = 'autoflag_closure_error_{0:d}'.format(i)
            def_fields = ','.join([pipeline.bpcal_id[i], pipeline.gcal_id[i], pipeline.target_id[i]])
            def_calfields = ','.join([pipeline.bpcal_id[i], pipeline.gcal_id[i]])
            if config['autoflag_closure_error'].get('fields', 'auto') != 'auto' and \
               not set(config['autoflag_closure_error'].get('fields', 'auto').split(',')) <= set(['gcal', 'bpcal', 'target']):
                raise KeyError("autoflag on phase fields can only be 'auto' or be a combination of 'gcal', 'bpcal' or 'target'")
            if config['autoflag_closure_error'].get('calibrator_fields', 'auto') != 'auto' and \
               not set(config['autoflag_closure_error'].get('calibrator_fields', 'auto')) <=  set(['gcal', 'bpcal']):
                raise KeyError("autoflag on phase calibrator fields can only be 'auto' or be a combination of 'gcal', 'bpcal'")

            fields = def_fields if config['autoflag_closure_error'].get('fields', 'auto') == 'auto' else \
                     ",".join([getattr(pipeline, key + "_id")[i] for key in config['autoflag_closure_error'].get('fields').split(',')])
            calfields = def_calfields if config['autoflag_closure_error'].get('calibrator_fields', 'auto') == 'auto' else \
                     ",".join([getattr(pipeline, key + "_id")[i] for key in config['autoflag_closure_error'].get('calibrator_fields').split(',')])

            recipe.add("cab/politsiyakat_cal_phase", step,
                {
                    "msname": msname,
                    "field": fields,
                    "cal_field": calfields,
                    "scan_to_scan_threshold": config["autoflag_closure_error"]["scan_to_scan_threshold"],
                    "baseline_to_group_threshold": config["autoflag_closure_error"]["baseline_to_group_threshold"],

                    "dpi": config['autoflag_closure_error'].get('dpi', 300),
                    "plot_size": config['autoflag_closure_error'].get('plot_size', 6),
                    "nproc_threads": config['autoflag_closure_error'].get('threads', 8),
                    "data_column": config['autoflag_closure_error'].get('column', "DATA")
                },
                input=pipeline.input, output=pipeline.output,
                label="{0:s}: Flag out baselines with closure errors")


        if pipeline.enable_task(config, 'flagging_summary'):
            step = 'flagging_summary_crosscal_{0:d}'.format(i)
            recipe.add('cab/casa_flagdata', step,
                {
                  "vis"         : msname,
                  "mode"        : 'summary',
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Flagging summary  ms={1:s}'.format(step, msname))



import sys
import warnings
import stimela.dismissable as sdm

NAME = 'Make HI Cube'
def worker(pipeline, recipe, config):
    mslist = ['{0:s}-{1:s}.ms'.format(did, config['label']) for did in pipeline.dataid]
    prefix = pipeline.prefix
    restfreq = config.get('restfreq','1.420405752GHz')
    npix = config.get('npix', 1024)
    cell = config.get('cell', 7)
    weight = config.get('weight', 'natural')
    robust = config.get('robust', 0)


    for i, msname in enumerate(mslist):
        if pipeline.enable_task(config, 'uvcontsub'):
            prefix = '{0:s}_{1:d}'.format(pipeline.prefix, i)
            step = 'contsub_{:d}'.format(i)
            recipe.add('cab/casa_uvcontsub', step, 
                {
                    "msname"    : msname,
                    "fitorder"  : config['uvcontsub'].get('fitorder', 1),
                    "fitspw"    : sdm.dismissable(config['uvcontsub'].get('fitspw',None))
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Subtract continuum'.format(step))

        if pipeline.enable_task(config, 'sunblocker'):
            if config['sunblocker']['use_contsub']:
                msname = msname+'.contsub'
            step = 'sunblocker_{0:d}'.format(i)
            recipe.add("cab/sunblocker", step, 
                {
                    "command"   : "phazer",
                    "inset"     : msname,
                    "outset"    : msname,
                    "imsize"    : config['sunblocker'].get('imsize', npix),
                    "cell"      : config['sunblocker'].get('cell', cell),
                    "pol"       : 'i',
                    "threshold" : config['sunblocker'].get('threshold', 4),
                    "mode"      : 'all',
                    "radrange"  : 0,
                    "angle"     : 0,
                    "show"      : prefix + '.sunblocker.pdf',
                    "verb"      : True,
                    "dryrun"    : True,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Block out sun'.format(step))

            
    if pipeline.enable_task(config, 'wsclean_image'):
        if config['wsclean_image']['use_contsub']:
            mslist = ['{0:s}-{1:s}.ms.contsub'.format(did, config['label']) for did in pipeline.dataid]
        step = 'wsclean_image_HI'
        spwid = config['wsclean_image'].get('spwid', 0)
        nchans = config['wsclean_image'].get('nchans','all')
        # Construct weight specification
        if config['wsclean_image'].get('weight', 'natural') == 'briggs':
            weight = 'briggs {0:.3f}'.format( config['wsclean_image'].get('robust', robust))
        else:
            weight = config['wsclean_image'].get('weight', weight)
        if nchans=='all': nchans=pipeline.nchans[0][spwid]
        recipe.add('cab/wsclean', step,
              {                       
                  "msname"    : mslist,
                  "weight"    : weight,
                  "npix"      : config['wsclean_image'].get('npix', npix),
                  "trim"      : dsm(config['wsclean_image'].get('trim', None)),
                  "scale"     : config['wsclean_image'].get('cell', cell),
                  "prefix"    : pipeline.prefix+'_HI',
                  "niter"     : config['wsclean_image'].get('niter', 1000000),
                  "mgain"     : config['wsclean_image'].get('mgain', 0.90),
                  "channelsout"     : nchans,
                  "auto-threshold"  : config['wsclean_image'].get('autothreshold', 5),
                  "auto-mask"  :   config['wsclean_image'].get('automask', 3),
                  "channelrange" : config['wsclean_image'].get('channelrange', [0, pipeline.nchans[0][spwid]]),
              },  
        input=pipeline.input,
        output=pipeline.output,
        label='{:s}:: Image HI'.format(step))

        if config['wsclean_image']['make_cube']:
            #nchans = config['wsclean_image'].get('nchans', pipeline.nchans[0][spwid])
            if not config['wsclean_image'].get('niter', 1000000): imagetype=['image','dirty']
            else: imagetype=['image','dirty','psf','residual','first-residual','model']
            for mm in imagetype:
                step = 'make_{0:s}_cube'.format(mm.replace('-','_'))
                recipe.add('cab/fitstool', step,
                    {    
                        "image"    : [pipeline.prefix+'_HI-{0:04d}-{1:s}.fits:output'.format(d,mm) for d in xrange(nchans)],
                        "output"   : pipeline.prefix+'_HI-{0:s}-cube.fits'.format(mm),
                        "stack"    : True,
                        "delete-files" : True,
                        "fits-axis": 'FREQ',
                    },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Make {1:s} cube from wsclean {1:s} channels'.format(step,mm.replace('-','_')))

    if pipeline.enable_task(config, 'casa_image'):
        if config['casa_image']['use_contsub']:
            mslist = ['{0:s}-{1:s}.ms.contsub'.format(did, config['label']) for did in pipeline.dataid]
        step = 'casa_image_HI'
        spwid = config['casa_image'].get('spwid', 0)
        nchans = config['casa_image'].get('nchans','all')
        if nchans=='all': nchans=pipeline.nchans[0][spwid]
        recipe.add('cab/casa_clean', step,
            {
                 "msname"         :    mslist,
                 "prefix"         :    pipeline.prefix+'_HI',
#                 "field"          :    target,
                 "mode"           :    'channel',
                 "nchan"          :    nchans,
                 "start"          :    config['casa_image'].get('startchan', 0,),
                 "interpolation"  :    'nearest',
                 "niter"          :    config['casa_image'].get('niter', 1000000),
                 "psfmode"        :    'hogbom',
                 "threshold"      :    config['casa_image'].get('threshold', '10mJy'),
                 "npix"           :    config['casa_image'].get('npix', npix),
                 "cellsize"       :    config['casa_image'].get('cell', cell),
                 "weight"         :    config['casa_image'].get('weight', weight),
                 "robust"         :    config['casa_image'].get('robust', robust),
#                 "wprojplanes"    :    1,
                 "port2fits"      :    True,
                 "restfreq"       :    restfreq,
            },
            input=pipeline.input,
            output=pipeline.output,
            label='{:s}:: Image HI'.format(step))


    if pipeline.enable_task(config, 'sofia'):
        if config['sofia']['imager']=='casa': cubename=pipeline.prefix+'_HI.image.fits:output'
        elif config['sofia']['imager']=='wsclean': cubename=pipeline.prefix+'_HI-cube.fits:output'
        step = 'sofia_sources'
        recipe.add('cab/sofia', step,
            {
            "import.inFile"         : cubename,
            "steps.doFlag"          : config['sofia'].get('flag', False),
            "steps.doScaleNoise"    : True,
            "steps.doSCfind"        : True,
            "steps.doMerge"         : config['sofia'].get('merge', True),
            "steps.doReliability"   : False,
            "steps.doParameterise"  : False,
            "steps.doWriteMask"     : True,
            "steps.doMom0"          : True,
            "steps.doMom1"          : False,
            "steps.doWriteCat"      : False,
            "flag.regions"          : config['sofia'].get('flagregion', []),
            "scaleNoise.statistic"  : config['sofia'].get('rmsMode', 'mad'),
            "SCfind.threshold"      : config['sofia'].get('threshold', 4),
            "SCfind.rmsMode"        : config['sofia'].get('rmsMode', 'mad'),
            "merge.radiusX"         : config['sofia'].get('mergeX', 2),
            "merge.radiusY"         : config['sofia'].get('mergeY', 2),
            "merge.radiusZ"         : config['sofia'].get('mergeZ', 3),
            "merge.minSizeX"        : config['sofia'].get('minSizeX', 3),
            "merge.minSizeY"        : config['sofia'].get('minSizeY', 3),
            "merge.minSizeZ"        : config['sofia'].get('minSizeZ', 5),
            },
            input=pipeline.input,
            output=pipeline.output,
            label='{0:s}:: Make SoFiA mask and images'.format(step))

    if pipeline.enable_task(config, 'flagging_summary'):
        for i,msname in enumerate(mslist):
            step = 'flagging_summary_image_HI_{0:d}'.format(i)
            recipe.add('cab/casa_flagdata', step,
                {
                  "vis"         : msname,
                  "mode"        : 'summary',
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Flagging summary  ms={1:s}'.format(step, msname))



import os
import sys

NAME = 'Split and average target data'

def worker(pipeline, recipe, config):
    label = config['label']
    for i in range(pipeline.nobs):
        msname = pipeline.msnames[i]
        target = pipeline.target[i]
        tms = '{0:s}-{1:s}.ms'.format(msname[:-3], label)
        flagv = tms + '.flagversions'
        prefix = pipeline.prefixes[i]

        if pipeline.enable_task(config, 'split_target'):
            step = 'split_target_{:d}'.format(i)
            if os.path.exists('{0:s}/{1:s}'.format(pipeline.msdir, tms)) or \
                   os.path.exists('{0:s}/{1:s}'.format(pipeline.msdir, flagv)):

                os.system('rm -rf {0:s}/{1:s} {0:s}/{2:s}'.format(pipeline.msdir, tms, flagv))

            recipe.add('cab/casa_split', step,
                {
                    "vis"           : msname,
                    "outputvis"     : tms,
                    "timebin"       : config['split_target'].get('time_average', ''),
                    "width"         : config['split_target'].get('freq_average', 1),
                    "datacolumn"    : 'corrected',
                    "correlation"   : config['split_target'].get('correlation', ''),
                    "field"         : str(target),
                    "keepflags"     : True,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Split and average data ms={1:s}'.format(step, msname))
            
        if pipeline.enable_task(config, 'prepms'):
            step = 'prepms_{:d}'.format(i)
            recipe.add('cab/msutils', step,
                {
                  "msname"  : tms,
                  "command" : 'prep' ,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Add BITFLAG column ms={1:s}'.format(step, tms))

        if pipeline.enable_task(config, 'obsinfo'):
            if config['obsinfo'].get('listobs', True):
                step = 'listobs_{:d}'.format(i)
                recipe.add('cab/casa_listobs', step,
                    {
                      "vis"         : tms,
                      "listfile"    : '{0:s}-{1:s}-obsinfo.txt'.format(prefix, label),
                      "overwrite"   : True,
                    },
                    input=pipeline.input,
                    output=pipeline.output,
                    label='{0:s}:: Get observation information ms={1:s}'.format(step, tms))
    
            if config['obsinfo'].get('summary_json', True):
                 step = 'summary_json_{:d}'.format(i)
                 recipe.add('cab/msutils', step,
                    {
                      "msname"      : tms,
                      "command"     : 'summary',
                      "outfile"     : '{0:s}-{1:s}-obsinfo.json'.format(prefix, label),
                    },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Get observation information as a json file ms={1:s}'.format(step, tms))



import argparse
import yaml
import meerkathi
import os
import copy
import ruamel.yaml
import itertools
from collections import OrderedDict

DEFAULT_CONFIG = os.path.join(meerkathi.pckgdir, 'default-config.yml')

def is_valid_file(parser, arg):
    if not os.path.exists(arg):
        parser.error("The file '%s' does not exist!" % arg)

    return arg


class config_parser:
    __ARGS = None
    __GROUPS = None

    @classmethod
    def __store_args(cls, args, arg_groups):
        """ Store arguments for later retrieval """

        if cls.__ARGS is not None:
            meerkathi.log.warn("Replacing existing stored arguments '{}'"
                               "with '{}'.", __ARGS, args)

        cls.__ARGS = args
        cls.__GROUPS = arg_groups

    @property
    def arg_groups(self):
        """ Retrieve groups """
        cls = self.__class__
        if cls.__GROUPS is None:
            raise ValueError("No arguments were stored. "
                             "Please call store_args first.")

        return copy.deepcopy(cls.__GROUPS)

    @property
    def args(self):
        """ Retrieve stored arguments """
        cls = self.__class__
        if cls.__ARGS is None:
            raise ValueError("No arguments were stored. "
                             "Please call store_args first.")

        return copy.deepcopy(cls.__ARGS)

    @classmethod
    def __primary_parser(cls, add_help=False):
        parser = argparse.ArgumentParser("MeerKATHI HI and Continuum Imaging Pipeline.\n"
                                         "(C) RARG, SKA-SA 2016-2017.\n"
                                         "All rights reserved.",
                                         add_help=add_help)
        add = parser.add_argument
        add("-v","--version", action='version',version='{0:s} version {1:s}'.format(parser.prog, meerkathi.__version__))
        add('-c', '--config',
            type=lambda a: is_valid_file(parser, a),
            default=DEFAULT_CONFIG,
            help='Pipeline configuration file (YAML/JSON format)')

        add('-gd', '--get-default',
            help='Name file where the configuration should be saved')

        add('-aaf', '--add-all-first', action='store_true',
            help='Add steps from all workers to pipeline before exucting. Default is execute each workers as they are encountered.')

        add('-bl', '--stimela-build',
            help='Label of stimela build to use',
            default=None)

        add('-pcs', '--print-calibrator-standard',
            help='Prints auxilary calibrator standard into the log',
            action='store_true')

        add('-wd', '--workers-directory', default='{:s}/workers'.format(meerkathi.pckgdir),
            help='Directory where pipeline workers can be found. These are stimela recipes describing the pipeline')

        add('-ce', '--config-editor', action='store_true',
            help='Start the interactive configuration editor (requires X session with decent [ie. firefox] webbrowser installed).')

        add('-rv', '--report-viewer', action='store_true',
            help='Start the interactive report viewer (requires X session with decent [ie. firefox] webbrowser installed).')

        add('--interactive-port', type=int, default=8888,
            help='Port on which to listen when an interactive mode is selected (e.g the configuration editor)')

        return parser

    __HAS_BEEN_INIT = False

    def __init__(self, args=None):
        """ Configuration parser. Sets up command line interface for MeerKATHI
            This is a singleton class, and should only be initialized once
        """
        cls = self.__class__
        if cls.__HAS_BEEN_INIT:
            return

        cls.__HAS_BEEN_INIT = True

        """ Extract """

        #=========================================================
        # Handle the configuration file argument first,
        # if one is supplied use that for defaulting arguments
        # created further down the line, otherwise use the
        # default configuration file
        #=========================================================
        # Create parser object
        parser  = cls.__primary_parser()

        # Lambda for transforming sections and options
        xformer = lambda s: s.replace('-', '_')
        def _str2bool(v):
            if v.upper() in ("YES","TRUE"):
                return True
            elif v.upper() in ("NO","FALSE"):
                return False
            else:
                raise argparse.ArgumentTypeError("Failed to convert argument. Must be one of "
                                                 "'yes', 'true', 'no' or 'false'.")

        def _nonetype(v):
            if v.upper() in ("NONE","NULL"):
                return None
            else:
                return str(v)

        def _subparser_tree(sections,
                            base_section="",
                            update_only = False,
                            args = None,
                            parser = None):
            """ Recursively creates subparser tree for the config """
            if sections is None:
                return

            sec_defaults = {xformer(k): v for k,v in sections.iteritems() }

            # Transform keys
            # Add subsection / update when necessary
            groups = OrderedDict()
            for opt, default in sec_defaults.iteritems():
                if opt == "__helpstr": continue
                option_name = base_section + "_" + opt if base_section != "" else opt
                if isinstance(default, dict):
                    groups[opt] = _subparser_tree(default,
                                                  base_section=option_name,
                                                  update_only=update_only,
                                                  args=args,
                                                  parser=parser)
                else:
                    if update_only:
                        parser.set_defaults(**{option_name: default})
                        groups[opt] = getattr(args, option_name)
                    else:
                        if isinstance(default, list):
                            parser.add_argument("--%s" % option_name,
                                                type=str,
                                                default=default,
                                                nargs="?",
                                                action='append')
                        elif isinstance(default, bool):
                            parser.add_argument("--%s" % option_name,
                                                type=_str2bool,
                                                default=default,
                                                nargs="?",
                                                const=True)
                        elif default is None:
                            parser.add_argument("--%s" % option_name,
                                                type=_nonetype,
                                                default=default,
                                                nargs="?")
                        else:
                            parser.add_argument("--%s" % option_name,
                                                type=type(default),
                                                default=default,
                                                nargs="?")

                        groups[opt] = default
            return groups

        # Parse user commandline options, loading defaults either from the default pipeline or user-supplied pipeline
        args_bak = copy.deepcopy(args)
        args, remainder = parser.parse_known_args(args_bak)
        if args.config:
            meerkathi.log.info("Loading defaults from user configuration '{}'".format(args.config))
            with open(args.config, 'r') as f:
                file_config = ruamel.yaml.load(f, ruamel.yaml.RoundTripLoader, version=(1,1))
        else:
            meerkathi.log.info("Loading defaults from installation configuration '{}'".format(DEFAULT_CONFIG))
            with open(DEFAULT_CONFIG, 'r') as f:
                file_config = ruamel.yaml.load(f, ruamel.yaml.RoundTripLoader, version=(1,1))
        parser = cls.__primary_parser(add_help=True)
        groups = _subparser_tree(file_config, parser=parser)
        args, remainder = parser.parse_known_args(args_bak)
        if len(remainder) > 0:
            raise RuntimeError("The following arguments were not parsed: %s" ",".join(remainder))
        groups = _subparser_tree(file_config, update_only=True, args=args, parser=parser)

        cls.__store_args(args, groups)

    @classmethod
    def log_options(cls):
        """ Prints argument tree to the logger for prosterity to behold """
        meerkathi.log.info("".join(["".ljust(25,"#"), " PIPELINE CONFIGURATION ", "".ljust(25,"#")]))
        def _tree_print(branch, indent="\t"):
            dicts = OrderedDict( [(k, v) for k, v in branch.iteritems() if isinstance(v, dict)] )
            other = OrderedDict( [(k, v) for k, v in branch.iteritems() if not isinstance(v, dict)] )

            def _printval(k, v):
                if isinstance(v, dict):
                    (indent=="\t") and meerkathi.log.info(indent.ljust(60,"#"))
                    meerkathi.log.info(indent + "Subsection %s:" % k)
                    (indent=="\t") and meerkathi.log.info(indent.ljust(60,"#"))
                    (indent!="\t") and meerkathi.log.info(indent.ljust(60,"-"))
                    _tree_print(v, indent=indent+"\t")
                else:
                    meerkathi.log.info("%s%s= %s" %(indent,
                                                   k.ljust(30),
                                                   v))

            for k, v in other.iteritems(): _printval(k, v)
            for k, v in dicts.iteritems(): _printval(k, v)
        ordered_groups = OrderedDict(sorted(cls.__GROUPS.items(),
                                             key=lambda p: p[1].get("order",0)))
        _tree_print(ordered_groups)
        meerkathi.log.info("".join(["".ljust(25,"#"), " END OF CONFIGURATION ", "".ljust(25,"#")]))



import os
import sys

NAME = "Conbine datasets"

def worker(pipeline, recipe, config):

    newid = config.get('newid')
    step = 'combine_{:s}'.format(newid)
    msnames = list(pipeline.msnames)
    msmeta = '{0:s}/{1:s}.json'.format(pipeline.data_path, pipeline.dataid[0])
    pipeline.init_names([newid])
    msname = pipeline.msnames[0]
    os.system('rm -fr {0:s}/{1:s}'.format(pipeline.msdir, msname))

    recipe.add('cab/casa_concat', step,
        {   
          "vis"         : msnames,
          "concatvis"   : msname,
        },  
        input=pipeline.input,
        output=pipeline.output,
        label='{0:s}:: Combine datasets ms={1:s}'.format(step, msname))

    os.system('cp {0:s} {1:s}/{2:s}.json'.format(msmeta, pipeline.data_path, pipeline.dataid[0]))



import meerkathi.dispatch_crew.catalog_parser as cp
import meerkathi
import os

__DB_FILENAME = os.path.join(meerkathi.pckgdir, "data/southern_calibrators.txt")

__CALIBRATOR_DB = None

def calibrator_database():
    """ Return the Southern standard calibrator database """

    global __CALIBRATOR_DB

    # Do a lazy load
    if __CALIBRATOR_DB is not None:
        return __CALIBRATOR_DB

    # OK its not loaded, read it in

    # There isn't a Southern standard in CASA
    # so construct a little database of them for reference
    meerkathi.log.info("Obtaining divine knowledge from: %s" % __DB_FILENAME)

    __CALIBRATOR_DB = cp.catalog_parser(__DB_FILENAME)

    return __CALIBRATOR_DB




import os
import sys
import subprocess
import itertools
import meerkathi
import meerkathi.dispatch_crew.meerkat_archive_interface as mai
import stimela.dismissable as sdm
import warnings

NAME = "Get convert and extract data"

def worker(pipeline, recipe, config):
    # this bit is intrinsic to meerkathi because it changes the pipeline input parameters... cannot be run in stimela
    if pipeline.enable_task(config, 'meerkat_query_available'):
        cfg = config["meerkat_query_available"]
        if cfg["poll_mode"] == "override":
            meerkathi.log.info("User specified MeerKAT filenames manually. Will now contact MeerKAT archive for their metadata...")
            dids = config["dataid"] if isinstance(config["dataid"], list) else [config["dataid"]]
            mdata = []
            pipeline.metadata = []
            for did in dids:
                mdata = \
                itertools.chain(mdata, mai.query_metadatas(pipeline.data_path,
                                                           cfg["query_url"],
                                                           filename="".join([did, ".h5"]),
                                                           product_type_name=None,
                                                           product_num_channels=None,
                                                           time_extent=None,
                                                           required_intents=[],
                                                           required_minimum_duration=0,
                                                           required_fields=[],
                                                           required_description=".*"))
            for m in mdata:
                bn = os.path.splitext(os.path.basename(m["Filename"]))[0]
                mai.dump_observation_metadata(pipeline.data_path,
                                              bn + '.json',
                                              m)
                pipeline.metadata.append('{0:s}/{1:s}.json'.format(pipeline.data_path, bn))
            pipeline.init_names(config["dataid"])
        elif cfg["poll_mode"] == "poll":
            meerkathi.log.info("Polling the MeerKAT archive for the latest imaging observations...")
            mdata = mai.query_metadatas(pipeline.data_path,
                                        cfg["query_url"],
                                        filename=None,
                                        product_type_name=cfg["product_type_name"],
                                        product_num_channels=cfg["product_num_channels"],
                                        time_extent=cfg["when_observed"],
                                        required_intents=cfg["required_intents"],
                                        required_minimum_duration=cfg["minimum_observation_duration"],
                                        required_fields=cfg["required_fields"],
                                        required_description=cfg["description_matches"])
            for m in mdata:
                bn = os.path.splitext(os.path.basename(m["Filename"]))[0]
                mai.dump_observation_metadata(pipeline.data_path,
                                              bn + '.json',
                                              m)
            pipeline.init_names([os.path.splitext(os.path.basename(m["Filename"]))[0] for m in mdata])

            if pipeline.nobs == 0:
                meerkathi.log.warn("The archive says there is no new data matching your search parameters. "
                                   "Try adjusting the search parameters. There is no work left to do.")
                sys.exit(0)
        else:
            raise ValueError("MK Archive poll_mode only accepts 'override' and 'poll'. Check your config!")
    else: # not going to query MeerKAT archive for h5 file metadata.... user specifying MS directly
        meerkathi.log.warn("User specified dataset names manually and chose not to query the archive. "
                           "If this is MeerKAT data you better have metadata downloaded already!")
        pipeline.init_names(config["dataid"])

    for i in range(pipeline.nobs):
        msname = pipeline.msnames[i]
        h5file = pipeline.h5files[i]
        basename = os.path.splitext(os.path.basename(h5file))[0]
        prefix = pipeline.prefixes[i]
        if isinstance(pipeline.data_path, list):
            data_path = pipeline.data_path[i]
        else:
            data_path = pipeline.data_path

        if pipeline.enable_task(config, 'download'):
            cfg = config['download']
            if isinstance(cfg["data_url"], list):
                data_url = cfg["data_url"][i]
            else:
                data_url = cfg["data_url"]
            dm = cfg['download_mode'].upper()
            if dm == "MEERKAT":
                step = 'download_{:d}'.format(i)
                def checked_download(h5file, basename):
                    (data_url != "") and meerkathi.log.warn("Download mode set to MeerKAT, but user specified a data-url. This will be ignored")
                    if not os.path.exists(os.path.join(pipeline.data_path, basename + '.json')): # user is being silly... download metadata anyway
                        mdata = mai.query_metadatas(pipeline.data_path,
                                            config["meerkat_query_available"]["query_url"],
                                            filename=h5file,
                                            product_type_name=None,
                                            product_num_channels=None,
                                            time_extent=None,
                                            required_intents=[],
                                            required_minimum_duration=0,
                                            required_fields=[],
                                            required_description=".*")

                        if len(mdata) == 0:
                            raise RuntimeError("Invalid product specified. You may want to run the meerkat_archive_query step or double check your dataid.")
                        meta = mdata[0]
                    else:
                        meta = mai.load_observation_metadata(pipeline.data_path,
                                                             basename + '.json')
                    if not mai.check_observation_cache(pipeline.data_path,
                                                       h5file,
                                                       meta):
                        mai.download_observations(pipeline.data_path, [meta])
                # add function to recipe
                recipe.add(checked_download, step, 
                      {
                       "h5file"    : h5file,
                       "basename"  : basename,
                      },
                    label='{0:s}:: Downloading MeerKAT data'.format(step))

            elif dm == "MANUAL":
                step = 'download_{:d}'.format(i)
                if os.path.exists('{0:s}/{1:s}'.format(pipeline.data_path, msname)) \
                    and not config['download'].get('reset', False):
                    meerkathi.log.warn('ms already exists, and reset is not enabled. Will attempt to resume')
                    recipe.add('cab/curl', step, {
                        "url"   : data_url,
                        "output": msname if config['download'].get('untar', False) else msname + '.tar',
                        "continue-at": "-"
                    },
                    input=pipeline.input,
                    output=pipeline.data_path,
                    label='{0:s}:: Downloading data'.format(step))
                else:
                    os.system('rm -rf {0:s}/{1:s}'.format(pipeline.data_path, msname))

                    recipe.add('cab/curl', step, {
                        "url"   : data_url,
                        "output": msname if config['download'].get('untar', False) else msname + '.tar',
                    },
                    input=pipeline.input,
                    output=pipeline.data_path,
                    label='{0:s}:: Downloading data'.format(step))


        if pipeline.enable_task(config, 'h5toms'):
            step = 'h5toms_{:d}'.format(i)

            if os.path.exists('{0:s}/{1:s}'.format(pipeline.msdir, msname)):
                os.system('rm -rf {0:s}/{1:s}'.format(pipeline.msdir, msname))

            recipe.add('cab/h5toms', step,
                {
                    "hdf5files"     : [h5file],
                    "output-ms"     : msname,
                    "no-auto"       : False,
                    "tar"           : True,
                    "model-data"    : True,
                    "channel-range" : sdm.dismissable(config['h5toms'].get('channel_range', None)),
                    "full-pol"      : config['h5toms'].get('full_pol', False),
                },
                input=data_path,
                output=pipeline.output,
                label='{0:s}:: Convert hd5file to MS. ms={1:s}'.format(step, msname))


    if pipeline.enable_task(config, 'combine'):
        step = 'combine_data'
        newid = config['combine']['newid']
        msnames = pipeline.msnames
        metadata = pipeline.metadata[0]
        pipeline.init_names([newid])
        msname = pipeline.msnames[0]
        pipeline.metada = [metadata]

        if config['combine'].get('reset', True):
            if os.path.exists('{0:s}/{1:s}'.format(pipeline.msdir, msname)):
                os.system('rm -rf {0:s}/{1:s}'.format(pipeline.msdir, msname))

            recipe.add('cab/casa_concat', step, 
                {
                    "vis"       : msnames,
                    "concatvis" : msname,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Combine datasets'.format(step))

        if config['combine'].get('tar', True):
            step = 'tar_{:d}'.format(i)
            tar_options = config['combine'].get('tar_options', 'cvf')
            # Function to untar Ms from .tar file
            def tar(ms):
                mspath = os.path.abspath(pipeline.msdir)
                subprocess.check_call(['tar', tar_options,
                    os.path.join(mspath, ms+'.tar'),
                    os.path.join(mspath, ms),
                    ])
            # add function to recipe
            recipe.add(tar, step, 
                 {
                  "ms"        : msname,
                 },
                 label='{0:s}:: Get MS from tarbal ms={1:s}'.format(step, msname))

    for i, msname in enumerate(pipeline.msnames):
        if pipeline.enable_task(config, 'untar'):
                step = 'untar_{:d}'.format(i)
                tar_options = config['untar'].get('tar_options', 'xvf')
                # Function to untar Ms from .tar file
                def untar(ms):
                    mspath = os.path.abspath(pipeline.msdir)
                    subprocess.check_call(['tar', tar_options,
                        os.path.join(mspath, ms+'.tar'),
                        '-C', mspath])
                # add function to recipe
                recipe.add(untar, step, 
                     {
                      "ms"        : msname,
                     },
                     label='{0:s}:: Get MS from tarbal ms={1:s}'.format(step, msname))



# -*- coding: utf-8 -*-

import sys
import os
import traceback

pckgdir = os.path.dirname(os.path.abspath(__file__))

import stimela
import glob
from meerkathi.dispatch_crew.config_parser import config_parser as cp

# Distutils standard  way to do version numbering
import pkg_resources
try:
    __version__ = pkg_resources.require("meerkathi")[0].version
except pkg_resources.DistributionNotFound:
    __version__ = "dev"

import logging
import traceback
import meerkathi.dispatch_crew.caltables as mkct
import meerkathi.scripts as scripts
from SimpleHTTPServer import SimpleHTTPRequestHandler
from BaseHTTPServer import HTTPServer
from multiprocessing import Process
import webbrowser
import base64
from urllib import urlencode
import ruamel.yaml
import json
from meerkathi.dispatch_crew.reporter import reporter as mrr

MEERKATHI_LOG = os.path.join(os.getcwd(), "meerkathi.log")

def create_logger():
    """ Create a console logger """
    log = logging.getLogger(__name__)
    cfmt = logging.Formatter(('%(name)s - %(asctime)s %(levelname)s - %(message)s'))
    log.setLevel(logging.DEBUG)
    filehandler = logging.FileHandler(MEERKATHI_LOG)
    filehandler.setFormatter(cfmt)
    log.addHandler(filehandler)
    log.setLevel(logging.INFO)

    console = logging.StreamHandler()
    console.setLevel(logging.INFO)
    console.setFormatter(cfmt)

    log.addHandler(console)

    return log

# Create the log object
log = create_logger()

class MeerKATHI(object):
    def __init__(self, config, workers_directory, 
            stimela_build=None, prefix=None, 
            add_all_first=False):

        self.config = config

        self.add_all_first = add_all_first

        self.msdir = self.config['general']['msdir']
        self.input = self.config['general']['input']
        self.output = self.config['general']['output']
        self.data_path = self.config['general']['data_path']
        self.workers_directory = workers_directory
        # Add workers to packages
        sys.path.append(self.workers_directory)
        self.workers = []

        for i, (name,opts) in enumerate(self.config.iteritems()):
            if name.find('general')>=0:
                continue
            order = opts.get('order', i+1)

            if name.find('__')>=0:
                worker = name.split('__')[0] + '_worker'
            else:
                worker = name + '_worker'

            self.workers.append((name, worker, order))

        self.workers = sorted(self.workers, key=lambda a: a[2])

        self.prefix = prefix or self.config['general']['prefix']
        self.stimela_build = stimela_build
        self.recipes = {}
        # Workers to skip
        self.skip = []
        # Initialize empty lists for ddids, leave this up to get data worker to define
        self.init_names([])

    def init_names(self, dataid):
        """ iniitalize names to be used throughout the pipeline and associated 
            general fields that must be propagated
        """
        self.dataid = dataid
        self.nobs = len(self.dataid)
        self.h5files = ['{:s}.h5'.format(dataid) for dataid in self.dataid]
        self.msnames = ['{:s}.ms'.format(os.path.basename(dataid)) for dataid in self.dataid]
        self.split_msnames = ['{:s}_split.ms'.format(os.path.basename(dataid)) for dataid in self.dataid]
        self.cal_msnames = ['{:s}_cal.ms'.format(os.path.basename(dataid)) for dataid in self.dataid]
        self.prefixes = ['meerkathi-{:s}'.format(os.path.basename(dataid)) for dataid in self.dataid]

        for item in 'input msdir output'.split():
            value = getattr(self, item, None)
            if value:
                setattr(self, item, value)

        for item in 'data_path reference_antenna fcal bpcal gcal target'.split():
            value = getattr(self, item, None)
            if value and len(value)==1:
                value = value*nobs
                setattr(self, item, value)

    def enable_task(self, config, task):
        a = config.get(task, False)
        if a:
            return a['enable']
        else:
            False

    def run_workers(self):
        """ Runs the  workers """
        for _name, _worker, i in self.workers:
            try:
                worker = __import__(_worker)
            except ImportError:
                traceback.print_exc()
                raise ImportError('Worker "{0:s}" could not be found at {1:s}'.format(_worker, self.workers_directory))

            config = self.config[_name]
            if config.get('enable', True) is False:
                self.skip.append(_worker)
                continue
            # Define stimela recipe instance for worker
            # Also change logger name to avoid duplication of logging info
            recipe = stimela.Recipe(worker.NAME, ms_dir=self.msdir, 
                               loggername='STIMELA-{:d}'.format(i), 
                               build_label=self.stimela_build)
            # Don't allow pipeline-wide resume
            # functionality
            os.system('rm -f {}'.format(recipe.resume_file))
            # Get recipe steps
            # 1st get correct section of config file
            worker.worker(self, recipe, config)
            # Save worker recipes for later execution
            # execute each worker after adding its steps

            if self.add_all_first:
                log.info("Adding worker {0:s} before running".format(_worker))
                self.recipes[_worker] = recipe
            else:
                log.info("Running worker {0:s}".format(_worker))
                recipe.run()

        # Execute all workers if they saved for later execution
        try:
            if self.add_all_first:
                for worker in self.workers:
                    if worker not in self.skip:
                        self.recipes[worker[1]].run()
        finally: # write reports even if the pipeline only runs partially
            reporter = mrr(self)
            reporter.generate_reports()



def main(argv):
    log.info("")

    log.info("███╗   ███╗███████╗███████╗██████╗ ██╗  ██╗ █████╗ ████████╗██╗  ██╗██╗")
    log.info("████╗ ████║██╔════╝██╔════╝██╔══██╗██║ ██╔╝██╔══██╗╚══██╔══╝██║  ██║██║")
    log.info("██╔████╔██║█████╗  █████╗  ██████╔╝█████╔╝ ███████║   ██║   ███████║██║")
    log.info("██║╚██╔╝██║██╔══╝  ██╔══╝  ██╔══██╗██╔═██╗ ██╔══██║   ██║   ██╔══██║██║")
    log.info("██║ ╚═╝ ██║███████╗███████╗██║  ██║██║  ██╗██║  ██║   ██║   ██║  ██║██║")
    log.info("╚═╝     ╚═╝╚══════╝╚══════╝╚═╝  ╚═╝╚═╝  ╚═╝╚═╝  ╚═╝   ╚═╝   ╚═╝  ╚═╝╚═╝")
    log.info("")
    # parse config file and set up command line argument override parser
    log.info("Module installed at: {0:s} (version {1:s})".format(pckgdir, str(__version__)))
    log.info("A logfile will be dumped here: {0:s}".format(MEERKATHI_LOG))
    log.info("")
    args = cp(argv).args
    arg_groups = cp(argv).arg_groups
    def __host():
        httpd = HTTPServer(("", port), hndl)
        os.chdir(web_dir)
        try:
            httpd.serve_forever()
        except KeyboardInterrupt, SystemExit:
            httpd.shutdown()

    # User requests default config => dump and exit
    if args.get_default:
        log.info("Dumping default configuration to {0:s} as requested. Goodbye!".format(args.get_default))
        os.system('cp {0:s}/default-config.yml {1:s}'.format(pckgdir, args.get_default))
        return
    elif args.config_editor:
        log.info("Entering interactive mode as requested: MeerKATHI configuration editor")
        port = args.interactive_port
        file_abs = args.config
        with file(file_abs, 'r') as f:
            cfg_txt = json.dumps(ruamel.yaml.load(f, ruamel.yaml.RoundTripLoader, version=(1,1)))
        web_dir = os.path.join(os.path.dirname(scripts.__file__), 'conf_helper')
        log.info("Starting HTTP web server, listening on port %d and hosting directory %s" %
                 (port, web_dir))
        log.info("Press Ctrl-C to exit")
        hndl = SimpleHTTPRequestHandler

        wt = Process(target = __host)
        try:
            wt.start()
            webbrowser.open("http://localhost:%d/index.html?%s" % (port, urlencode({"filetxt":cfg_txt})))
            wt.join()
        except (KeyboardInterrupt, SystemExit):
            log.info("Interrupt received - shutting down web server. Goodbye!")
        return
    elif args.report_viewer:
        log.info("Entering interactive mode as requested: MEERKATHI report viewer")
        port = args.interactive_port
        web_dir = os.path.abspath(os.path.join(args.general_output, 'reports'))
        if not os.path.exists(web_dir):
            log.error("Reports directory '%s' does not yet exist. Has the pipeline been run here?" % web_dir)
            return
        log.info("Starting HTTP web server, listening on port %d and hosting directory %s" %
                 (port, web_dir))
        log.info("Press Ctrl-C to exit")
        hndl = SimpleHTTPRequestHandler

        wt = Process(target = __host)
        try:
            wt.start()
            webbrowser.open("http://localhost:%d/" % port)
            wt.join()
        except KeyboardInterrupt, SystemExit:
            log.info("Interrupt received - shutting down web server. Goodbye!")
        return

    # Very good idea to print user options into the log before running:
    cp().log_options()

    # Obtain some divine knowledge
    cdb = mkct.calibrator_database()

    if args.print_calibrator_standard:
        meerkathi.log.info("Found the following reference calibrators (in CASA format):")
        meerkathi.log.info(cdb)

    import exceptions
    try:
        pipeline = MeerKATHI(arg_groups,
                             args.workers_directory, stimela_build=args.stimela_build,
                             add_all_first=args.add_all_first, prefix=args.general_prefix)

        pipeline.run_workers()
    except exceptions.SystemExit as e:
        if e.code != 0:
            log.error("One or more pipeline workers enacted E.M.E.R.G.E.N.C.Y protocol {0:d} shutdown. This is likely a bug, please report.".format(e.code))
            log.error("Your logfile is here: {0:s}. You are running version: {1:s}".format(MEERKATHI_LOG, str(__version__)))
            sys.exit(1) #indicate failure
        else:
            log.info("One or more pipeline workers requested graceful shutdown. Goodbye!")
    except:
        log.error("Whoops... big explosion - you sent pipes flying all over the show! Time to call in the monkeywrenchers.")
        log.error("Your logfile is here: {0:s}. You are running version: {1:s}".format(MEERKATHI_LOG, str(__version__)))
        tb = traceback.format_exc()
        log.error(tb)
        sys.exit(1) #indicate failure



import meerkathi
import os
import json
import copy
import re
import nbconvert
import nbformat
from nbconvert import HTMLExporter
from nbconvert.preprocessors import ExecutePreprocessor
from nbconvert.preprocessors import CellExecutionError

REPORT_TEMPLATE=os.path.join(meerkathi.scripts.__path__[0], "obs_report", "Observation Report.ipynb")

class reporter:
    def __init__(self, pipeline):
        """ Process and dump a static html report for each ms in the pipeline """
        self.__ms = copy.deepcopy(pipeline.msnames)
        self.__outputdir = pipeline.output
        with open(REPORT_TEMPLATE) as f:
            self.__rep_template = f.read()
        self.__report_dir = os.path.join(self.__outputdir, "reports")
        if not os.path.exists(self.__report_dir):
            os.mkdir(self.__report_dir)

    def generate_reports(self):
        """ generate an html report for every ms in the pipeline """
        report_names = [os.path.join(self.__report_dir, "%s.report.html" % ms) for ms in self.__ms]
        for ms, rep in zip(self.__ms, report_names):
            meerkathi.log.info("Creating a report for dataset id '%s'. "
                               "The report will be dumped here: '%s'." % (ms, rep))
            # grab a fresh template
            ms_rep = nbformat.reads(self.__rep_template, as_version=4)

            def __customize(s):
                s = re.sub(r'msname\s*=\s*\S*',
                           'msname = \'%s\'' % os.path.splitext(os.path.basename(ms))[0],
                           s)
                s = re.sub(r'outputdir\s*=\s*\S*',
                           'outputdir = \'%s\'' % os.path.abspath(self.__outputdir),
                           s)
                return s

            # modify template to add the output directory and ms name
            ms_rep.cells[0]['source'] = '\n'.join(map(__customize, ms_rep.cells[0]['source'].split('\n')))

            # roll
            ep = ExecutePreprocessor(timeout=None, kernel_name='python2')
            try:
                ep.preprocess(ms_rep, {'metadata': {'path': os.path.abspath(self.__outputdir)}})
            except CellExecutionError: # reporting error is non-fatal
                out = None
                msg = 'Error executing the notebook for "%s".\n\n' % ms
                msg += 'See notebook "%s" for the traceback.' % rep
                meerkathi.log.error(msg)
            finally:
                #export to static HTML
                html_exporter = HTMLExporter()
                #html_exporter.template_file = 'basic'
                (body, resources) = html_exporter.from_notebook_node(ms_rep)
                with open(rep, 'w+') as f:
                    f.write(body)
                    #nbformat.write(ms_rep, f)



import collections
import itertools
import json
import os
import sys
import meerkathi
import pysolr
import requests
import progressbar
import re

def __standard_observation_query(filename=None,
                                 product_type_name="MeerKATAR1TelescopeProduct",
                                 product_num_channels=4096,
                                 time_extent="[NOW-3DAYS TO NOW]"):
    """ Constructs SOLR query based on arguments
        Specify None to remove criteria
        Returns SOLR query string
    """
    # Create a quick field+query type
    fq = collections.namedtuple('field_query', 'field query')

    query_list = [
        # Product name if specified
        fq('Filename', filename),
        # Only want MeerKAT AR1 Telescope Products
        fq('CAS.ProductTypeName', product_type_name),
        # Only want 4K mode
        fq('NumFreqChannels', product_num_channels),
        # Observations from the last 3 days
        fq('StartTime', time_extent)]

    # Construct the query
    query = ' AND '.join('%s:%s' % (fq.field, fq.query) for fq in query_list if fq.query is not None)
    return query

def __query_filter(solr_url,
                    query=None,
                    required_intents=['gaincal', 'bpcal', 'target'],
                    required_minimum_duration=2,
                    required_fields=[],
                    required_description='.*'):
    """
    Find recent telescope observations suitable for imaging
    """

    ONE_MINUTE = 60
    ONE_HOUR = 60 * ONE_MINUTE

    search = query if query is not None else __standard_observation_query()


    meerkathi.log.info("Querying solr server '%s' "
                        "with query string '%s'." % (solr_url, search))

    archive = pysolr.Solr(solr_url)

    def _observation_filter(solr_result):
        """
        Filter out KAT observations that don't cut the mustard
        for imaging purposes.
        """

        # Create an observation string for logging
        filename, description = (solr_result.get(f,'') for f in
            ('Filename', 'Description'))
        observation = '{} {}'.format(filename, description[:50])

        # Its only worth imaging observations with
        # gain and bandpass calibrations
        targets = solr_result.get('KatpointTargets', [])
        gaincals = sum('gaincal' in s for s in targets)
        bandpasses = sum('bpcal' in s for s in targets)
        obs_targets = sum('target' in s for s in targets)

        if ('gaincal' in required_intents) and gaincals == 0:
            meerkathi.log.warn('Ignoring "{}", no gain calibrators '
                                'are present.'.format(observation))
            return False

        if ('bpcal' in required_intents) and bandpasses == 0:
            meerkathi.log.warn('Ignoring "{}", no band pass calibrators '
                                'are present.'.format(observation))
            return False

        if ('target' in required_intents) and obs_targets == 0:
            meerkathi.log.warn('Ignoring "{}", no target fields '
                                'are present.'.format(observation))
            return False

        # Check that the description matches
        prod_desc = solr_result.get('Description', '')
        if not re.match(required_description, prod_desc):
            meerkathi.log.warn('Ignoring "{}", observation description doesn\'t match '
                               'user query Regex.'.format(observation))
            return False

        # Check that required fields have been observed
        obs_fields = solr_result.get('Targets', [])
        for f in required_fields:
            if f not in obs_fields:
                meerkathi.log.warn('Ignoring "{}", observation doesn\'t contain '
                                   'field "{}" as required by user.'.format(observation,
                                                                            f))
                return False


        # Don't take anything less than 2 hours in duration
        duration = solr_result['Duration']

        if duration < required_minimum_duration * ONE_HOUR:
            meerkathi.log.warn('Ignoring "{}", observation is '
                'only "{}"s long, need to be at minimum "{}"s.'.format(observation,
                                                                       duration,
                                                                       required_minimum_duration * ONE_HOUR))
            return False

        return True
    hits = []
    curr_cursor = "*"
    res = archive.search(search, sort='ProductName desc, id asc', rows=1000, cursorMark='*')
    # Step through query results - may be many pages
    while curr_cursor != res.nextCursorMark:
        curr_cursor = res.nextCursorMark
        hits += filter(_observation_filter, res)
        res = archive.search(search, sort='ProductName desc, id asc', rows=1000, cursorMark=curr_cursor)

    return hits

def load_observation_metadata(directory, filename):
    """ Load observation metadata """

    with open(os.path.join(directory, filename), 'r') as f:
        return json.load(f)

def dump_observation_metadata(directory, filename, observation):
    """ Dump observation metadata """

    with open(os.path.join(directory, filename), 'w') as f:
        return json.dump(observation, f)

def query_metadatas(input_dir,
                    solr_url,
                    filename=None,
                    product_type_name="MeerKATAR1TelescopeProduct",
                    product_num_channels=4096,
                    time_extent="[NOW-3DAYS TO NOW]",
                    required_intents=['gaincal','bpcal','target'],
                    required_minimum_duration=2,
                    required_fields=[],
                    required_description='.*'):
    """ Return a list of observation metadatas """

    # If a specific HDF5 file is specified, attempt to load local metadata first
    results = []
    if filename is not None:
        if isinstance(filename, str):
            lstfiles = [filename]
        elif isinstance(filename, list):
            lstfiles = filename
        else:
            raise ValueError("Expected filename list")
        for f in lstfiles:
            query = __standard_observation_query(filename=f,
                                                 product_type_name=product_type_name,
                                                 product_num_channels=product_num_channels,
                                                 time_extent=time_extent)

            basename = os.path.splitext(f)[0]
            metadata_filename = ''.join([basename, '.json'])
            metadata_path = os.path.join(input_dir, metadata_filename)

            # If it exists, load and return the metadata as our observation object
            if os.path.exists(metadata_path):
                meerkathi.log.info("Observation metadata exists locally at '{}'. "
                                    "Reusing it.".format(metadata_path))

                results = itertools.chain(results, [load_observation_metadata(input_dir,
                                                                              metadata_filename)])
            else:
                # Nope, need to download it from the server
                results = itertools.chain(results, __query_filter(solr_url,
                                                                  query,
                                                                  required_intents,
                                                                  required_minimum_duration,
                                                                  required_fields,
                                                                  required_description))

    else:
        query = __standard_observation_query(filename=None,
                                             product_type_name=product_type_name,
                                             product_num_channels=product_num_channels,
                                             time_extent=time_extent)

        results = itertools.chain(results, __query_filter(solr_url,
                                                          query,
                                                          required_intents,
                                                          required_minimum_duration,
                                                          required_fields,
                                                          required_description))
    return list(results)

def download_observations(directory, observations):
    """ Download the specified observations
        args:
        : observations: list of observation metadata
        returns: list of filenames downloaded
    """

    ONE_KB = 1024
    ONE_MB = ONE_KB*ONE_KB
    filenames = []
    for observation in observations:
        targets = observation.get('KatpointTargets', [])
        gaincals = sum('gaincal' in s for s in targets)
        bandpasses = sum('bpcal' in s for s in targets)
        obs_targets = sum('target' in s for s in targets)

        meerkathi.log.info('%s -- %-2d -- %-2d-- %-2d-- %s -- %s -- %s' % (
            observation['Filename'], gaincals, bandpasses, obs_targets,
            observation['StartTime'], observation['Description'], observation['Duration']))

        # Infer the HTTP location from the KAT archive file location
        location = observation['FileLocation'][0]
        location = location.replace('/var/kat', 'http://kat-archive.kat.ac.za', 1)
        filename = observation['Filename']
        url = os.path.join(location, filename)

        filename = os.path.join(directory, filename)
        file_exists = os.path.exists(filename) and os.path.isfile(filename)
        local_file_size = os.path.getsize(filename) if file_exists else 0
        headers = { "Range" : "bytes={}-".format(local_file_size) }

        r = requests.get(url, headers=headers, stream=True)

        # Server doesn't care about range requests and is just
        # sending the entire file
        if r.status_code == 200:
            meerkathi.log.info("Downloading '{}'")
            remote_file_size = r.headers.get('Content-Length', None)
            file_exists = False
            local_file_size = 0
        elif r.status_code == 206:
            if local_file_size > 0:
                meerkathi.log.info("'{}' already exists, "
                    "resuming download from {}.".format(
                        filename, local_file_size))

            # Create a fake range if none exists
            fake_range = "{}-{}/{}".format(local_file_size, sys.maxint,
                sys.maxint - local_file_size)

            remote_file_size = r.headers.get('Content-Range', fake_range)
            remote_file_size = remote_file_size.split('/')[-1]
        elif r.status_code == 416:
            meerkathi.log.info("'{}' already downloaded".format(filename))
            remote_file_size = local_file_size
            return filename
        else:
            raise ValueError("HTTP Error Code {}".format(r.status_code))

        meerkathi.log.info('%s %s %s' % (url, remote_file_size, r.status_code))

        f = open(filename, 'ab' if file_exists else 'wb')
        bar = (progressbar.ProgressBar(max_value=progressbar.UnknownLength)
            if remote_file_size is None else progressbar.ProgressBar(
                maxval=int(remote_file_size)))

        #Download chunks of file and write to disk
        try:
            with f, bar:
                downloaded = local_file_size
                for chunk in r.iter_content(chunk_size=ONE_MB):
                    if chunk:
                        f.write(chunk)
                        downloaded += len(chunk)
                        bar.update(downloaded)
        except KeyboardInterrupt as kbe:
            pass
            #log.warn("Quitting download on Keyboard Interrupt")
        filenames.append(filename)

    return filename

def check_observation_cache(directory, h5filename, observation):
    """
    Return True if both an h5file and the associated observation metadata
    exist. The file size indicated in the metadata must agree with that
    of the h5file.
    """

    # Does the h5 file exist?
    h5file = os.path.join(directory, h5filename)

    if not os.path.exists(h5file):
        return False

    # Compare metadata file size vs h5 file size
    h5_size = os.path.getsize(h5file)
    obs_size = observation["FileSize"][0]

    if not obs_size == os.path.getsize(h5file):
        meerkathi.log.warn("'{}' file size '{}' "
            "differs from that in the observation metadata '{}'."
                .format(h5filename, h5_size, obs_size))
        return False

    return True



import os
import sys

NAME = "Prepare data for calibration"

def worker(pipeline, recipe, config):

    for i in range(pipeline.nobs):

        msname = pipeline.msnames[i]
        prefix = pipeline.prefixes[i]
        
        if pipeline.enable_task(config, 'fixvis'):
            step = 'fixvis_{:d}'.format(i)
            recipe.add('cab/casa_fixvis', step,
                {
                    "vis"        : msname,
                    "reuse"      : False,
                    "outputvis"  : msname,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Fix UVW coordinates ms={1:s}'.format(step, msname))

        if pipeline.enable_task(config, 'prepms'):
            step = 'prepms_{:d}'.format(i)
            recipe.add('cab/msutils', step,
                {
                  "msname"  : msname,
                  "command" : 'prep' ,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Add BITFLAG column ms={1:s}'.format(step, msname))

        if pipeline.enable_task(config, 'add_spectral_weights'):
            step = 'estimate_weights_{:d}'.format(i)
            recipe.add('cab/msutils', step,
                {
                  "msname"          : msname,
                  "command"         : 'estimate_weights',
                  "stats_data"      : config['add_spectral_weights'].get('stats_data', 'use_package_meerkat_spec'),
                  "weight_columns"  : config['add_spectral_weights'].get('weight_columns', ['WEIGHT', 'WEIGHT_SPECTRUM']),
                  "noise_columns"   : config['add_spectral_weights'].get('noise_columns', ['SIGMA', 'SIGMA_SPECTRUM']),
                  "write_to_ms"     : config['add_spectral_weights'].get('write_to_ms', True),
                  "plot_stats"      : prefix + '-noise_weights.png',
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Adding Spectral weights using MeerKAT noise specs ms={1:s}'.format(step, msname))



import sys
import os

NAME = 'Inspect data'

def worker(pipeline, recipe, config):

    def get_field(field):
        if field in ['bpcal', 'gcal', 'target', 'fcal']:
            name = getattr(pipeline, field)[i]
        else:
            name = field
        return str(name)

    uvrange = config.get('uvrange', '')
    for i in range(pipeline.nobs):
        msname = pipeline.msnames[i]
        prefix = pipeline.prefixes[i]
        label = config.get('label', '')

        if pipeline.enable_task(config, 'real_imag'):
            fields = config['real_imag'].get('fields', 'gcal,bpcal').split(',')
            for field_ in fields:
                for col in ['baseline', 'scan']:
                    field = get_field(field_)
                    step = 'plot_real_imag_{0:d}'.format(i)
                    recipe.add('cab/casa_plotms', step,
                       {
                        "vis"           : msname,
                        "field"         : field,
                        "correlation"   : 'XX,YY',
                        "timerange"     : '',
                        "antenna"       : '',
                        "xaxis"         : 'imag',
                        "xdatacolumn"   : config['real_imag'].get('datacolumn', 'corrected'),
                        "yaxis"         : 'real',
                        "ydatacolumn"   : config['real_imag'].get('datacolumn', 'corrected'),
                        "avgtime"       : config['real_imag'].get('avgtime', ''),
                        "avgchannel"    : config['real_imag'].get('avgchannel', ''),
                        "coloraxis"     : col,
                        "iteraxis"      : 'corr',
                        "plotfile"      : '{0:s}-{1:s}-{2:s}-{3:s}-reim.png'.format(prefix, label, field_, col),
                        "expformat"     : 'png',
                        "exprange"      : 'all',
                        "overwrite"     : True,
                        "showgui"       : False,
                        "uvrange"       : uvrange,
                       },
                       input=pipeline.input,
                       output=pipeline.output,
                       label='{0:s}:: Plot imag vs real for field {1:s} ms={2:s} col={3:s}'.format(step, field, msname, col))

        if pipeline.enable_task(config, 'amp_phase'):
            fields = config['amp_phase'].get('fields', 'gcal,bpcal').split(',')
            for field_ in fields:
                for col in ['baseline', 'scan']:
                    field = get_field(field_)
                    step = 'plot_amp_phase_{0:d}'.format(i)
                    recipe.add('cab/casa_plotms', step,
                       {
                        "vis"           : msname,
                        "field"         : field,
                        "correlation"   : 'XX,YY',
                        "timerange"     : '',
                        "antenna"       : '',
                        "xaxis"         : 'phase',
                        "xdatacolumn"   : config['amp_phase'].get('datacolumn', 'corrected'),
                        "yaxis"         : 'amp',
                        "ydatacolumn"   : config['amp_phase'].get('datacolumn', 'corrected'),
                        "avgtime"       : config['amp_phase'].get('avgtime', ''),
                        "avgchannel"    : config['amp_phase'].get('avgchannel', ''),
                        "coloraxis"     : col,
                        "iteraxis"      : 'corr',
                        "plotfile"      : '{0:s}-{1:s}-{2:s}-{3:s}-ap.png'.format(prefix, label, field_, col),
                        "expformat"     : 'png',
                        "exprange"      : 'all',
                        "overwrite"     : True,
                        "showgui"       : False,
                        "uvrange"       : uvrange,
                       },
                       input=pipeline.input,
                       output=pipeline.output,
                       label='{0:s}:: Plot amp vs phase for field {1:s} ms={2:s} col={3:s}'.format(step, field, msname, col))

        if pipeline.enable_task(config, 'amp_uvwave'):
            fields = config['amp_uvwave'].get('fields', 'gcal,bpcal').split(',')
            for field_ in fields:
                field = get_field(field_)
                step = 'plot_uvwave_{0:d}'.format(i)
                recipe.add('cab/casa_plotms', step,
                   {
                    "vis"           : msname,
                    "field"         : field,
                    "correlation"   : 'XX,YY',
                    "timerange"     : '',
                    "antenna"       : '',
                    "xaxis"         : 'uvwave',
                    "xdatacolumn"   : config['amp_uvwave'].get('datacolumn', 'corrected'),
                    "yaxis"         : 'amp',
                    "ydatacolumn"   : config['amp_uvwave'].get('datacolumn', 'corrected'),
                    "avgtime"       : config['amp_uvwave'].get('avgtime', ''),
                    "avgchannel"    : config['amp_uvwave'].get('avgchannel', ''),
                    "coloraxis"     : 'baseline',
                    "iteraxis"      : 'corr',
                    "expformat"     : 'png',
                    "exprange"      : 'all',
                    "plotfile"      : '{0:s}-{1:s}-{2:s}-ampuvwave.png'.format(prefix, label, field_),
                    "overwrite"     : True,
                    "showgui"       : False,
                    "uvrange"       : uvrange,
                   },
                   input=pipeline.input,
                   output=pipeline.output,
                   label='{0:s}:: Plot uv-wave for field {1:s} ms={2:s}'.format(step, field, msname))

        if pipeline.enable_task(config, 'amp_ant'):
            fields = config['amp_ant'].get('fields', 'gcal,bpcal').split(',')
            for field_ in fields:
                field = get_field(field_)
                step = 'plot_uvwave_{0:d}'.format(i)
                recipe.add('cab/casa_plotms', step,
                   {
                    "vis"           : msname,
                    "field"         : field,
                    "correlation"   : 'XX,YY',
                    "timerange"     : '',
                    "antenna"       : '',
                    "xaxis"         : 'antenna1',
                    "xdatacolumn"   : config['amp_ant'].get('datacolumn', 'corrected'),
                    "yaxis"         : 'amp',
                    "ydatacolumn"   : config['amp_ant'].get('datacolumn', 'corrected'),
                    "avgtime"       : config['amp_ant'].get('avgtime', ''),
                    "avgchannel"    : config['amp_ant'].get('avgchannel', ''),
                    "coloraxis"     : 'corr',
                    "expformat"     : 'png',
                    "exprange"      : 'all',
                    "plotfile"      : '{0:s}-{1:s}-{2:s}-ampant.png'.format(prefix, label, field_),
                    "overwrite"     : True,
                    "showgui"       : False,
                    "uvrange"       : uvrange,
                   },
                   input=pipeline.input,
                   output=pipeline.output,
                   label='{0:s}:: Plot ampant for field {1:s} ms={2:s}'.format(step, field, msname))


        if pipeline.enable_task(config, 'phase_uvwave'):
            fields = config['phase_uvwave'].get('fields', 'gcal,bpcal').split(',')
            for field_ in fields:
                field = get_field(field_)
                step = 'phase_uvwave_{0:d}'.format(i)
                recipe.add('cab/casa_plotms', step,
                   {
                    "vis"           : msname,
                    "field"         : field,
                    "correlation"   : 'XX,YY',
                    "timerange"     : '',
                    "antenna"       : '',
                    "xaxis"         : 'uvwave',
                    "xdatacolumn"   : config['phase_uvwave'].get('datacolumn', 'corrected'),
                    "yaxis"         : 'phase',
                    "ydatacolumn"   : config['phase_uvwave'].get('datacolumn', 'corrected'),
                    "avgtime"       : config['phase_uvwave'].get('avgtime', ''),
                    "avgchannel"    : config['phase_uvwave'].get('avgchannel', ''),
                    "coloraxis"     : 'baseline',
                    "iteraxis"      : 'corr',
                    "expformat"     : 'png',
                    "exprange"      : 'all',
                    "plotfile"      : '{0:s}-{1:s}-{2:s}-phaseuvwave.png'.format(prefix, label, field_),
                    "overwrite"     : True,
                    "showgui"       : False,
                    "uvrange"       : uvrange,
                   },
                   input=pipeline.input,
                   output=pipeline.output,
                   label='{0:s}:: Plot phase uv-wave for field {1:s} ms={2:s}'.format(step, field, msname))


        if pipeline.enable_task(config, 'amp_scan'):
            fields = config['amp_scan'].get('fields', 'gcal,target,bpcal').split(',')
            for field_ in fields:
                field = get_field(field_)
                step = 'plot_ampscan_{0:d}'.format(i)
                recipe.add('cab/casa_plotms', step,
                   {
                    "vis"           : msname,
                    "field"         : field,
                    "correlation"   : 'XX,YY',
                    "timerange"     : '',
                    "antenna"       : '',
                    "xaxis"         : 'scan',
                    "xdatacolumn"   : config['amp_scan'].get('datacolumn', 'corrected'),
                    "yaxis"         : 'amp',
                    "ydatacolumn"   : config['amp_scan'].get('datacolumn', 'corrected'),
                    "avgtime"       : config['amp_scan'].get('avgtime', ''),
                    "avgchannel"    : config['amp_scan'].get('avgchannel', ''),
                    "coloraxis"     : 'baseline',
                    "iteraxis"      : 'corr',
                    "expformat"     : 'png',
                    "exprange"      : 'all',
                    "plotfile"      : '{0:s}-{1:s}-{2:s}-ampscan.png'.format(prefix, label, field_),
                    "overwrite"     : True,
                    "showgui"       : False,
                    "uvrange"       : uvrange,
                   },
                   input=pipeline.input,
                   output=pipeline.output,
                   label='{0:s}:: Plot amp_v_scan for field {1:s} ms={2:s}'.format(step, field, msname))



import re
from scipy.optimize import curve_fit
import numpy as np
import copy

class catalog_parser:
    def __init__(self, filename):
        """
            The all-knowning catalog class
            Give me a filename and I shall pass on divine knowledge
        """
        cls = self.__class__
        self._cat = cls.read_caltable(filename)

    @property
    def db(self):
        """ Returns a copy of divine sky knowledge """
        return copy.deepcopy(self._cat)

    def __str__(self):
      """ Return multiline string describing the calibrator database """
      lines = [""]
      lines.extend(["\t%s\tEpoch:%d\tRA:%3.2f\tDEC:%3.2f\t"
                    "S_v0:%.4f\tv0:%.4e\ta:%.4f\tb:%.4f\tc:%.4f\td:%.4f" %
                  (str(name).ljust(15),
                   db["epoch"],
                   db["ra"],
                   db["decl"],
                   db["S_v0"],
                   db["v0"],
                   db["a_casa"],
                   db["b_casa"],
                   db["c_casa"],
                   db["d_casa"])
            for name, db in self._cat.iteritems()])
      return '\n'.join(lines)

    @classmethod
    def read_caltable(cls, filename):
        """
        Read calibrator database (specified in MHz)
        and returns a dictionary containing the following
        :filename: filename of caltable database
        :returns: for every source (name = key):
                    Epoch, RA, Declination,
                    a_ghz, b_ghz, c_ghz, d_ghz,
                    a_mhz, b_mhz, c_mhz, d_mhz,
                    S_v0, a_casa, b_casa, c_casa, d_casa, v0
        :side-effects: none
        """
        calibrator_db = {}
        with open(filename) as f:
            line=f.readline()
            ln_no = 1
            while line:
                #discard comments
                command = line.split("//")[0]

                #empty line ?
                if command.strip() == "":
                    line = f.readline()
                    ln_no += 1
                    continue
                cmd = None
                #source ?
                valset = re.match(r"^name=(?P<name>[0-9A-Za-z\-+_ ]+)[ ]+"
                                  r"epoch=(?P<epoch>[0-9]+)[ ]+"
                                  r"ra=(?P<ra>[+\-]?[0-9]+h[0-9]+m[0-9]+(?:.[0-9]+)?s)[ ]+"
                                  r"dec=(?P<decl>[+\-]?[0-9]+d[0-9]+m[0-9]+(?:.[0-9]+)?s)[ ]+"
                                  r"a=(?P<a>[+\-]?[0-9]+(?:.[0-9]+)?)[ ]+"
                                  r"b=(?P<b>[+\-]?[0-9]+(?:.[0-9]+)?)[ ]+"
                                  r"c=(?P<c>[+\-]?[0-9]+(?:.[0-9]+)?)[ ]+"
                                  r"d=(?P<d>[+\-]?[0-9]+(?:.[0-9]+)?)$",
                                  command)
                #else alias ?
                if not valset:
                    valset = re.match(r"^alias src=(?P<src>[0-9A-Za-z\-+_ ]+)[ ]+"
                                      r"dest=(?P<dest>[0-9A-Za-z\-+_ ]+)$",
                                      command)
                    #else illegal
                    if not valset:
                        raise RuntimeError("Illegal line encountered while parsing"
                                           "southern standard at line %d:'%s'" %
                                           (ln_no, line))

                    cmd = "alias"
                else:
                    cmd = "add"

                if cmd == "add":
                    # parse sources (spectra in MHz)
                    name = valset.group("name").strip()
                    epoch = int(valset.group("epoch"))
                    ra = valset.group("ra")
                    valset_ra = re.match(r"^(?P<h>[+\-]?[0-9]+)h"
                                         r"(?P<m>[0-9]+)m"
                                         r"(?P<s>[0-9]+(?:.[0-9]+)?)s$",
                                         ra)
                    ra = np.deg2rad((float(valset_ra.group("h")) +
                                     float(valset_ra.group("m")) / 60.0 +
                                     float(valset_ra.group("s")) / 3600) / 24.0 * 360)
                    decl = valset.group("decl")
                    valset_decl = re.match(r"^(?P<d>[+\-]?[0-9]+)d"
                                           r"(?P<m>[0-9]+)m"
                                           r"(?P<s>[0-9]+(?:.[0-9]+)?)s$",
                                           decl)
                    decl = np.deg2rad(float(valset_decl.group("d")) + \
                                      float(valset_decl.group("m")) + \
                                      float(valset_decl.group("s")))

                    a = float(valset.group("a"))
                    b = float(valset.group("b"))
                    c = float(valset.group("c"))
                    d = float(valset.group("d"))

                    # convert models to Perley Butler GHz format
                    k = np.log10(1000)
                    ag = a + (b * k) + (c * k ** 2) + (d * k ** 3)
                    bg = b + (2 * c * k) + (3 * d * k ** 2)
                    cg = c + (3 * d * k)
                    dg = d

                    # convert model components to CASA/MT format
                    s_v0, a_casa, b_casa, c_casa, d_casa = cls.convert_pb_to_casaspi(0.8, 1.8, 1.4,
                                                                                     ag, bg, cg, dg)
                    calibrator_db[name] = {"epoch": epoch, "ra": ra, "decl": decl,
                                           "a_ghz": ag, "b_ghz": bg, "c_ghz": cg, "d_ghz": dg,
                                           "a_mhz": a, "b_mhz": b, "c_mhz": c, "d_mhz": d,
                                           "S_v0": s_v0,
                                           "a_casa": a_casa, "b_casa": b_casa,
                                           "c_casa": c_casa, "d_casa": d_casa,
                                           "v0": 1.4e9}
                elif cmd == "alias":
                    src = valset.group("src")
                    dest = valset.group("dest")
                    if not src in calibrator_db:
                        raise RuntimeError("%s has not been defined. Cannot alias "
                                           "%s to %s in line %d" %
                                           (src,dest,src,ln_no))
                    calibrator_db[dest] = calibrator_db[src]

                # finally parse next line
                line = f.readline()
                ln_no += 1

            return calibrator_db

    @classmethod
    def convert_pb_to_casaspi(cls, vlower, vupper, v0, a, b, c, d):
        """
        Coverts between the different conventions:
        PB: 10 ** [a + b * log10(v) + c * log10(v) ** 2 + d * log10(v) ** 3]
        CASA/Meqtrees SPI: S(v0) * (v/v0) ** [a' + b'*ln(v/v0) + c'*ln(v/v0) ** 2 + d'*ln(v/v0) ** 3]

        args:
        :vlower, vupper: range (same unit as a, b, c, d coefficients!) to fit for a',b',c',d'
        :v0: reference frequency (same unit as vlower, vupper!)
        :a,b,c,d: PB coefficients (for the unit used in vlower, vupper and v0!)

        side-effects: none
        """
        if vlower > vupper:
            raise ValueError("vlower must be lower than vupper")

        def pbspi(v, a, b, c, d):
            return 10 ** (a + b * np.log10(v) + c * np.log10(v) ** 2 + d * np.log10(v) ** 3)
        def casaspi(v, v0, I, a, b, c, d):
            return I * (v/v0) ** (a + b * np.log(v/v0) + c * np.log(v/v0) ** 2 + d * np.log(v/v0) ** 3)

        I = pbspi(v0, a, b, c, d)

        v = np.linspace(vlower, vupper, 10000)
        popt, pcov = curve_fit(lambda v, a, b, c, d: casaspi(v, v0, I, a, b, c, d), v, pbspi(v,a,b,c,d))
        perr = np.sqrt(np.diag(pcov))
        assert np.all(perr < 1.0e-6)

        # returns (S(v0), a', b', c', d')
        return I, popt[0], popt[1], popt[2], popt[3]



import ruamel.yaml
import math
import numpy
import yaml
import meerkathi
import meerkathi.dispatch_crew.caltables as mkct 
import re
import astropy.io.fits as fitsio
import codecs

def angular_dist_pos_angle (ra1, dec1, ra2, dec2):
    """Computes the angular distance between the two points on a sphere, and
    the position angle (North through East) of the direction from 1 to 2."""

    # Knicked from ska-sa/tigger
    ra = ra2 - ra1
    sind0, sind, cosd0, cosd = numpy.sin(dec1), numpy.sin(dec2), numpy.cos(dec1), numpy.cos(dec2)
    sina, cosa = numpy.sin(ra)*cosd, numpy.cos(ra)*cosd
    x = cosa*sind0 - sind*cosd0
    y = sina
    z = cosa*cosd0 + sind*sind0
    PA = numpy.arctan2(y,-x)
    R = numpy.arccos(z)

    return R,PA

def categorize_fields(msinfo):
    with open(msinfo, 'r') as f:
        info = ruamel.yaml.load(f, ruamel.yaml.RoundTripLoader)

    names = info['FIELD']['NAME']
    ids = info['FIELD']['SOURCE_ID']
    intents = info['FIELD']['INTENTS']
    intent_ids = info['FIELD']['STATE_ID']

    mapping = {
        'fcal' : (['CALIBRATE_FLUX'], []),
        'gcal' : (['CALIBRATE_AMPL', 'CALIBRATE_PHASE'], []),
        'bpcal': (['CALIBRATE_BANDPASS'], []),
        'target': (['TARGET'], []),
    }

    for i, field in enumerate(names):
        ints = intents[intent_ids[i]].split(',')
        for intent in ints:
            for ftype in mapping:
                if intent in mapping[ftype][0]:
                    mapping[ftype][-1].append(field)

    return mapping

def get_field_id(msinfo, field_name):
    """ Gets field id """
    with open(msinfo, 'r') as f:
        info = ruamel.yaml.load(f, ruamel.yaml.RoundTripLoader)
    names = info['FIELD']['NAME']
    ids = info['FIELD']['SOURCE_ID']
    if field_name not in names:
        raise KeyError("Could not find field '%s'" % field_name)
    return ids[names.index(field_name)]

def select_gcal(msinfo, targets, calibrators, mode='nearest'):
    """
      Automatically select gain calibrator
    """
    with open(msinfo, 'r') as f:
        info = ruamel.yaml.load(f, ruamel.yaml.RoundTripLoader)

    names = info['FIELD']['NAME']
    ids = info['FIELD']['SOURCE_ID']
    dirs = info['FIELD']['REFERENCE_DIR']

    def index(field):
        if isinstance(field, str):
            idx = names.index(field)
        elif isinstance(field, int):
            idx = ids.index(field)
        return idx

    if mode == 'most_scans':
        most_scans = 0
        gcal = None
        for fid in calibrators:
            idx = index(fid)
            field = str(ids(idx))
            if most_scans < len(info['SCAN'][field]):
                most_scans = len(info['SCAN'][field])
                gcal = names[idx]
    elif mode == 'nearest':
        tras = []
        tdecs = []
        for target in targets:
            idx = index(target)

            tras.append(dirs[idx][0][0])
            tdecs.append(dirs[idx][0][1])

        mean_ra = numpy.mean(tras)
        mean_dec = numpy.mean(tdecs)
        
        nearest_dist = numpy.inf
        gcal = None
        for field in calibrators:
            idx = index(field)
            ra = dirs[idx][0][0]
            dec = dirs[idx][0][1]
            distance = angular_dist_pos_angle(mean_ra, mean_dec, ra, dec)[0]
            if nearest_dist > distance:
                nearest_dist = distance
                gcal = names[idx]

    return gcal

def observed_longest(msinfo, bpcals):
    """
      Automatically select bandpass calibrator
    """
    with open(msinfo, 'r') as f:
        info = yaml.load(f)

    names = info['FIELD']['NAME']
    ids = info['FIELD']['SOURCE_ID']
    dirs = info['FIELD']['REFERENCE_DIR']

    def index(field):
        if isinstance(field, str):
            idx = names.index(field)
        elif isinstance(field, int):
            idx = ids.index(field)
        return idx

    most_time = 0
    field = None
    for bpcal in bpcals:
        idx = index(bpcal)
        bpcal = str(ids[idx])
        total_time = numpy.sum(info['SCAN'][bpcal].values())
        if total_time > most_time:
            most_time = total_time
            field = names[idx]

    return field

def field_observation_length(msinfo, field):
    with open(msinfo, 'r') as f:
        info = yaml.load(f)
    
    names = info['FIELD']['NAME']
    ids = info['FIELD']['SOURCE_ID']
    
    def index(field):
        if isinstance(field, str):
            idx = names.index(field)
        elif isinstance(field, int):
            idx = ids.index(field)
        return idx
    field = str(ids[index(field)])

    return  numpy.sum(info['SCAN'][field].values())


def find_in_native_calibrators(msinfo, field):
    """Check if field is in the South Calibrators database. 
       Return model if it is. Else, return False. 
    """

    db = mkct.calibrator_database()
    if field not in db.db.keys():
        return False

    with open(msinfo, 'r') as stdr:
        info = yaml.load(stdr)

    ref = info['SPW']['REF_FREQUENCY'][0] # Centre frequency of first channel
    bw = info['SPW']['TOTAL_BANDWIDTH'][0]
    nchan = info['SPW']['NUM_CHAN'][0]
    
    src = db.db[field]
    aghz = src["a_casa"]
    bghz = src["b_casa"]
    cghz = src["c_casa"]
    dghz = src["d_casa"]

    return dict(I=src['S_v0'], 
	a=src['a_casa'], 
	b=src['b_casa'], 
	c=src['c_casa'], 
	d=src['d_casa'], 
	ref=src['v0'])

def meerkat_refant(obsinfo):
    """ get reference antenna. Only works for MeerKAT observations downloaded through meerkathi"""

    with open(obsinfo) as stdr:
        info = yaml.load(stdr)
    return info['RefAntenna']


def find_in_casa_calibrators(msinfo, field):
    """Check if field is in the CASA NRAO Calibrators database. 
       Return model if it is. Else, return False. 
    """
   
    with open(meerkathi.pckgdir +'/data/casa_calibrators.yml') as stdr:
        db = yaml.load(stdr)
    
    found = False
    for src in db['models'].values():
        if field == src['3C']:
            found = True
            standards = src['standards']
            break
        else:
            _field = re.findall(r'\d+', field)
            if len(_field) == 2:
                for name in [src[d] for d in ['B1950', 'J2000', 'ALT']]:
                    _name = re.findall(r'\d+', name)
                    if ''.join(_field) == ''.join(_name) or ''.join(_field[:-1]) == ''.join(_name[:-1]):
                        found = True
                        standards = src['standards']
                        break
    if found:
        standard = standards.split(',')[0]
        return db['standards'][int(standard)]
    else:
        return False

def estimate_solints(msinfo, skymodel, Tsys_eta, dish_diameter, npol, gain_tol=0.05, j=3, save=False):
    if isinstance(skymodel, str):
        skymodel = [skymodel]
    flux = 0
    for name in skymodel:
        with fitsio.open(name) as hdu:
            model = hdu[1].data
        # Get total flux from model
        flux += model['Total_flux'].sum()

    # Get number of antennas
    with open(msinfo) as yr:
        info = yaml.load(yr)
    nant = len(info['ANT']['NAME'])

    # Get time and frequency resoltion of data
    dtime = info['EXPOSURE']
    bw = sum(info['SPW']['TOTAL_BANDWIDTH'])
    nchans = sum(info['SPW']['NUM_CHAN'])
    dfreq = bw/nchans
    
    k_b =  1.38e-23 # Boltzman's constant
    Jy = 1e-26 # 1 Jansky

    # estimate noise needed for a gain error of 'gain_tol' using Sandeep Sirothia's Equation (priv comm). 
    visnoise = flux * numpy.sqrt(nant - j) * gain_tol
    # calculate dt*df (solution intervals) needed to get that noise
    effective_area = numpy.pi * (dish_diameter / 2.0)**2
    dt_dfreq = ( 2 * k_b * Tsys_eta  / (Jy * numpy.sqrt(npol) * effective_area * visnoise) )**2 

    # return/save dt*df and the time, frequency resolution of the data
    if save:
        with codecs.open(msinfo, 'w', 'utf8') as yw:
            info['DTDF'] = dt_dfreq
            yaml.dump(data, yw, default_flow_style=False)
    
    return dt_dfreq, dtime, dfreq


def imaging_params(msinfo, spwid=0):
    with open(msinfo) as yr:
        info = yaml.load(yr)

    maxbl = info['MAXBL']
    dish_size = numpy.mean(info['ANTENNA']['DISH_DIAMETER'])
    freq = info['SPW']["REF_FREQUENCY"][spwid]
    wavelegnth = 2.998e8/freq
    
    FoV = numpy.rad2deg( 1.22 * wavelength / dish_size  )
    max_res = numpy.rad2deg( wavelength / maxbl )
    
    return max_res, FoV



import os
import sys
import yaml
import meerkathi
import stimela.dismissable as sdm
from meerkathi.dispatch_crew import utils

NAME = 'Self calibration loop'

CUBICAL_OUT = {
    "CORR_DATA"    : 'sc',
    "CORR_RES"     : 'sr',
}

CUBICAL_MT = {
    "Gain2x2"      : 'complex-2x2', 
    "GainDiag"      : 'complex-2x2',  #TODO:: Change this. Ask cubical to support this mode
    "GainDiagPhase": 'phase-diag',
}


def worker(pipeline, recipe, config):
    npix = config['img_npix']
    trim = config['img_trim']
    spwid = config.get('spwid', 0)
    cleanborder = 0
    cell = config['img_cell']
    mgain = config['img_mgain']
    niter = config['img_niter']
    auto_thresh = config['img_auto_threshold']
    auto_mask = config['img_auto_mask']
    robust = config['img_robust']
    nchans = config['img_nchans']
    pol = config.get('img_pol', 'I')
    thresh_pix = config['sf_thresh_pix']
    thresh_isl = config['sf_thresh_isl']
    column = config['img_column']
    joinchannels = config['img_joinchannels']
    fit_spectral_pol = config['img_fit_spectral_pol']
    gsols = config.get('cal_Gsols', [1,0])
    bsols = config.get('cal_Bsols', [0,1])
    taper = config.get('img_uvtaper', None)
    label = config['label']
    bjones = config.get('cal_Bjones', False)
    time_chunk = config.get('cal_time_chunk', 128)
    ncpu = config.get('ncpu', 9)
    mfsprefix = ["",'-MFS'][int(nchans>1)]

    mslist = ['{0:s}-{1:s}.ms'.format(did, label) for did in pipeline.dataid]
    prefix = pipeline.prefix

    # Define image() extract_sources() calibrate()
    # functions for convience

    def image(num):
        key = 'image_{}'.format(num)
        mask = False
        if config[key].get('peak_based_mask_on_dirty', False):
            mask = True
            step = 'image_{}_dirty'.format(num)
            recipe.add('cab/wsclean', step,
                  {                   
                      "msname"    : mslist,
                      "column"    : config[key].get('column', column),
                      "weight"    : 'briggs {}'.format(config.get('robust', robust)),
                      "npix"      : config[key].get('npix', npix),
                      "trim"      : config[key].get('trim', trim),
                      "scale"     : config[key].get('cell', cell),
                      "pol"       : config[key].get('pol', pol),
                      "channelsout"   : nchans,
                      "taper-gaussian" : sdm.dismissable(config[key].get('uvtaper', taper)),
                      "prefix"    : '{0:s}_{1:d}'.format(prefix, num),
                  },
            input=pipeline.input,
            output=pipeline.output,
            label='{:s}:: Make dirty image to create clean mask'.format(step))
            
            step = 'mask_dirty_{}'.format(num)
            recipe.add('cab/cleanmask', step,
               {
                 "image"           :  '{0:s}_{1:d}{2:s}-image.fits:output'.format(prefix, num, mfsprefix),
                 "output"          :  '{0:s}_{1:d}-mask.fits'.format(prefix, num),
                 "dilate"          :  False,
                 "peak-fraction"   :  0.5,
                 "no-negative"     :  True,
                 "boxes"           :  1,
                 "log-level"       :  'DEBUG',
               },
               input=pipeline.input,
               output=pipeline.output,
               label='{0:s}:: Make mask based on peak of dirty image'.format(step))
        
        elif config[key].get('mask', False):
            mask = True
            sigma = config[key].get('mask_sigma', None)
            pf = config[key].get('mask_peak_fraction', None)
            step = 'mask_{}'.format(num)
            recipe.add('cab/cleanmask', step,
               {
                 "image"           :  '{0:s}_{1:d}{2:s}-image.fits:output'.format(prefix, num-1, mfsprefix),
                 "output"          :  '{0:s}_{1:d}-mask.fits'.format(prefix, num),
                 "dilate"          :  False,
                 "peak-fraction"   :  sdm.dismissable(pf),
                 "sigma"           :  sdm.dismissable(sigma),
                 "no-negative"     :  True,
                 "boxes"           :  1,
                 "log-level"       :  'DEBUG',
               },
               input=pipeline.input,
               output=pipeline.output,
               label='{0:s}:: Make mask based on peak of dirty image'.format(step))

        step = 'image_{}'.format(num)
        image_opts = {                   
                  "msname"    : mslist,
                  "column"    : config[key].get('column', column),
                  "weight"    : 'briggs {}'.format(config[key].get('robust', robust)),
                  "npix"      : config[key].get('npix', npix),
                  "trim"      : config[key].get('trim', trim),
                  "scale"     : config[key].get('cell', cell),
                  "prefix"    : '{0:s}_{1:d}'.format(prefix, num),
                  "niter"     : config[key].get('niter', niter),
                  "mgain"     : config[key].get('mgain', mgain),
                  "pol"       : config[key].get('pol', pol),
                  "taper-gaussian" : sdm.dismissable(config[key].get('uvtaper', taper)),
                  "channelsout"     : nchans,
                  "joinchannels"    : config[key].get('joinchannels', joinchannels),
                  "fit-spectral-pol": config[key].get('fit_spectral_pol', fit_spectral_pol),
                  "auto-threshold": config[key].get('auto_threshold', auto_thresh),
              }
        if mask:
            image_opts.update( {"fitsmask" : '{0:s}_{1:d}-mask.fits:output'.format(prefix, num)} )
        else:
            image_opts.update( {"auto-mask" : config[key].get('auto_mask', auto_mask)} )

        recipe.add('cab/wsclean', step,
        image_opts,
        input=pipeline.input,
        output=pipeline.output,
        label='{:s}:: Make image after first round of calibration'.format(step))

    def make_cube(num, imtype='model'): 
        im = '{0:s}_{1}-cube.fits:output'.format(prefix, num)
        step = 'makecube_{}'.format(num)
        images = ['{0:s}_{1}-{2:04d}-{3:s}.fits:output'.format(prefix, num, i, imtype) for i in range(nchans)]
        recipe.add('cab/fitstool', step,
            {                   
                "image"     : images,
                "output"    : im,
                "stack"     : True,
                "fits-axis" : 'FREQ',
            },
            input=pipeline.input,
            output=pipeline.output,
            label='{0:s}:: Make convolved model'.format(step))

        return im

    def extract_sources(num):
        key = 'extract_sources_{0:d}'.format(num)
        if config[key].get('detection_image', False):
            step = 'detection_image_{0:d}'.format(num)
            detection_image = prefix + '-detection_image_{0:d}.fits:output'.format(num)
            recipe.add('cab/fitstool', step,
                {                   
                    "image"    : [prefix+'_{0:d}{2:s}-{1:s}.fits:output'.format(num, im, mfsprefix) for im in ('image','residual')],
                    "output"   : detection_image,
                    "diff"     : True,
                    "force"    : True,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Make convolved model'.format(step))
        else:
            detection_image = None

        spi_do = config[key].get('spi', False) 
        if spi_do:
            im = make_cube(num, 'image')
        else:
            im = '{0:s}_{1:d}{2:s}-image.fits:output'.format(prefix, num, mfsprefix)

        step = 'extract_{0:d}'.format(num)
        calmodel = '{0:s}_{1:d}-pybdsm'.format(prefix, num)
        if detection_image:
            blank_limit = 1e-9
        else:
            blank_limit = None
        recipe.add('cab/pybdsm', step,
            {                   
                "image"         : im,
                "thresh_pix"    : config[key].get('thresh_pix', thresh_pix),
                "thresh_isl"    : config[key].get('thresh_isl', thresh_isl),
                "outfile"       : '{:s}.fits:output'.format(calmodel),
                "blank_limit"   : sdm.dismissable(blank_limit),
                "adaptive_rms_box" : True,
                "port2tigger"   : True,
                "multi_chan_beam": spi_do,
                "spectralindex_do": spi_do,
                "detection_image": sdm.dismissable(detection_image),
            },
            input=pipeline.input,
            output=pipeline.output,
            label='{0:s}:: Extract sources'.format(step))

    def predict_from_fits(num, model, index):
        if isinstance(model, str) and len(model.split('+'))==2:
            combine = True
            mm = model.split('+')
            # Combine FITS models if more than one is given
            step = 'combine_models_' + '_'.join(map(str, mm))
            calmodel = '{0:s}_{1:d}-FITS-combined.fits:output'.format(prefix, num)
            cubes = [ make_cube(n, 'model') for n in mm]
            recipe.add('cab/fitstool', step,
                {
                    "image"    : cubes,
                    "output"   : calmodel,
                    "sum"      : True,
                    "force"    : True,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Add clean components'.format(step))
        else:
            calmodel = make_cube(num)
            
        step = 'predict_fromfits_{}'.format(num)
        recipe.add('cab/lwimager', 'predict', {
                "msname"        : mslist[index], 
                "simulate_fits" : calmodel,
                "column"        : 'MODEL_DATA',
                "img_nchan"     : nchans,
                "img_chanstep"  : 1,
                "nchan"         : pipeline.nchans[index][spwid],
                "cellsize"      : cell,
                "chanstep"      : 1,
        },
            input=pipeline.input,
            output=pipeline.output,
            label='{0:s}:: Predict from FITS ms={1:s}'.format(step, mslist[index]))


    def combine_models(models, num):
        model_names = ['{0:s}_{1:s}-pybdsm.lsm.html:output'.format(prefix, m) for m in models]
        model_names_fits = ['{0:s}/{1:s}_{2:s}-pybdsm.fits'.format(pipeline.output, prefix, m) for m in models]
        calmodel = '{0:s}_{1:d}-pybdsm-combined.lsm.html:output'.format(prefix, num)

        step = 'combine_models_' + '_'.join(map(str, models))
        recipe.add('cab/tigger_convert', step,
            {                   
                "input-skymodel"    : model_names[0],
                "append"    : model_names[1],
                "output-skymodel"   : calmodel,
                "rename"  : True,
                "force"   : True,
            },
            input=pipeline.input,
            output=pipeline.output,
            label='{0:s}:: Combined models'.format(step))

        return calmodel, model_names_fits

    def autoset_calibration_intervals(recipe, skymodel, num, key):
        ## No way around it. The recipe has to be executed at this point to get the sky model
        recipe.run()
        # Empty job que after execution
        recipe.jobs = []

        solints = []
        # skymodel = os.path.join(pipeline.output, skymodel.split(':')[0])
        for i in range(pipeline.nobs):
            msinfo = '{0:s}/{1:s}-{2:s}-obsinfo.json'.format(pipeline.output, pipeline.prefixes[i], label)
            phase_only = config[key].get('gain_matrix_type', 'GainDiagPhase') == 'GainDiagPhase'

            dtdf, dt, df = utils.estimate_solints(msinfo, skymodel, 
                  pipeline.Tsys_eta, pipeline.dish_diameter, 
                  npol=2, gain_tol=config[key].get('gain_tol', 0.05), 
                  j=2 if phase_only else 3)

            with open(msinfo) as yr:
                info = yaml.load(yr)
            nchans = sum( info['SPW']['NUM_CHAN'])
            target = info['FIELD']['NAME'].index(pipeline.target[i])
            
            tot_time = sum( info['SCAN'][str(target)].values() )

            # Minimum time bins needed for gain_error=gain_tol for entire bandwidth
            min_time_bin = int( (dtdf / (df * nchans)) / dt + 1)
            # Minimum freq bins needed for gain_error=gain_tol for entire time
            min_freq_bin = int( (dtdf / (tot_time)) / df + 1)

            gsols = [min_time_bin, 0]
            bsols = [0, min_freq_bin]
            solints.append([gsols,bsols])

        meerkathi.log.info('Product of time and frequency solution intervals is {0:.4g} [s . Hz]'.format(dtdf))
        
        return solints

    def calibrate_meqtrees(num):
        key = 'calibrate_{0:d}'.format(num)
        model = config[key].get('model', num)
        vismodel = config[key].get('add_vis_model', False)

        if config[key].get('visonly', False):
            vismodel = True
            calmodel = '{0:s}_{1:d}-nullmodel.txt'.format(prefix, num)
            with open(os.path.join(pipeline.input, calmodel), 'w') as stdw:
                stdw.write('#format: ra_d dec_d i\n')
                stdw.write('0.0 -30.0 1e-99')
            for i,msname in enumerate(mslist):
                predict_from_fits(num, model, i)
        else: 
            if isinstance(model, str) and len(model.split('+'))>1:
                combine = True
                mm = model.split('+')
                calmodel, fits_model = combine_models(mm, num)
            else:
                model = int(model)
                calmodel = '{0:s}_{1:d}-pybdsm.lsm.html:output'.format(prefix, model)
                fits_model = '{0:s}/{1:s}_{2:d}-pybdsm.fits'.format(pipeline.output, prefix, model)

        autosols = [],[]
        autosols_set = False
        if config[key].get('Gsols', gsols) == 'auto' or \
                       config[key].get('Bsols', gsols) == 'auto':
            autosols = autoset_calibration_intervals(recipe, fits_model, num, key)
            config[key]['Bjones'] = True
            autosols_set = True

        for i,msname in enumerate(mslist):
            if autosols_set:
                gsols_ = autosols[i][0] 
                bsols_ = autosols[i][1] 
            else:
                gsols_ = config[key].get('Gsols', gsols)
                bsols_ = config[key].get('Bsols', bsols)

            step = 'calibrate_{0:d}_{1:d}'.format(num, i)
            recipe.add('cab/calibrator', step,
               {
                 "skymodel"             : calmodel,
                 "add-vis-model"        : vismodel,
                 "msname"               : msname,
                 "threads"              : ncpu,
                 "column"               : "DATA",
                 "output-data"          : config[key].get('output_data', 'CORR_RES'),
                 "output-column"        : "CORRECTED_DATA",
                 "prefix"               : '{0:s}-{1:d}_meqtrees'.format(pipeline.dataid[i], num),
                 "label"                : 'cal{0:d}'.format(num),
                 "read-flags-from-ms"   : True,
                 "read-flagsets"        : "-stefcal",
                 "write-flagset"        : "stefcal",
                 "write-flagset-policy" : "replace",
                 "Gjones"               : True,
                 "Gjones-solution-intervals" : gsols_,
                 "Gjones-matrix-type"   : config[key].get('gain_matrix_type', 'GainDiag'),
                 "Gjones-ampl-clipping"      : True,
                 "Gjones-ampl-clipping-low"  : config.get('cal_gain_amplitude_clip_low', 0.5),
                 "Gjones-ampl-clipping-high" : config.get('cal_gain_amplitude_clip_high', 1.5),
                 "Bjones"                    : config[key].get('Bjones', False),
                 "Bjones-ampl-clipping"      : True,
                 "Bjones-solution-intervals" : bsols_,
                 "Bjones-ampl-clipping"      : config[key].get('Bjones', bjones),
                 "Bjones-ampl-clipping-low"  : config.get('cal_gain_amplitude_clip_low', 0.5),
                 "Bjones-ampl-clipping-high" : config.get('cal_gain_amplitude_clip_high', 1.5),
                 "make-plots"           : True,
                 "tile-size"            : time_chunk,
               },
               input=pipeline.input,
               output=pipeline.output,
               label="{0:s}:: Calibrate step {1:d} ms={2:s}".format(step, num, msname))

    def calibrate_cubical(num):
        key = 'calibrate_{0:d}'.format(num)
        model = config[key].get('model', num)
        if isinstance(model, str) and len(model.split('+'))>1:
            combine = True
            mm = model.split('+')
            calmodel, fits_model = combine_models(mm, num)
        else:
            model = int(model)
            calmodel = '{0:s}_{1:d}-pybdsm.lsm.html:output'.format(prefix, model)
            fits_model = '{0:s}/{1:s}_{2:d}-pybdsm.fits'.format(pipeline.output, prefix, model)

        autosols = [],[]
        autosols_set = False
        if config[key].get('Gsols', gsols) == 'auto' or \
                       config[key].get('Bsols', gsols) == 'auto':
            autosols = autoset_calibration_intervals(recipe, fits_model, num, key)
            config[key]['Bjones'] = True
            autosols_set = True

        if config[key].get('Bjones', bjones):
            jones_chain = 'G,B'
        else:
            jones_chain = 'G' 

        for i,msname in enumerate(mslist):
            if autosols_set:
                gsols_ = autosols[i][0] 
                bsols_ = autosols[i][1] 
            else:
                gsols_ = config[key].get('Gsols', gsols)
                bsols_ = config[key].get('Bsols', bsols)

            step = 'calibrate_cubical_{0:d}_{1:d}'.format(num, i)
            recipe.add('cab/cubical', step, 
                {   
                    "data-ms"          : msname, 
                    "data-column"      : 'DATA',
                    "model-column"     : 'MODEL_DATA' if config[key].get('add_vis_model', False) else ' "" ',
                    "j2-term-iters"    : 200,
                    "data-time-chunk"  : time_chunk,
                    "sel-ddid"         : sdm.dismissable(config[key].get('spwid', None)),
                    "dist-ncpu"        : ncpu,
                    "sol-jones"        : jones_chain,
                    "model-lsm"        : calmodel,
                    "out-name"         : '{0:s}-{1:d}_cubical'.format(pipeline.dataid[i], num),
                    "out-mode"         : CUBICAL_OUT[config[key].get('output_data', 'CORR_DATA')],
                    "out-plots-show"   : False,
                    "weight-column"    : config[key].get('weight_column', 'WEIGHT'),
                    "montblanc-dtype"  : 'float',
                    "j1-solvable"      : True,
                    "j1-type"          : CUBICAL_MT[config[key].get('gain_matrix_type', 'Gain2x2')],
                    "j1-time-int"      : gsols_[0],
                    "j1-freq-int"      : gsols_[1],
                    "j1-clip-low"      : config.get('cal_gain_amplitude_clip_low', 0.5),
                    "j1-clip-high"     : config.get('cal_gain_amplitude_clip_high', 1.5),
                    "j2-solvable"      : config[key].get('Bjones', bjones),
                    "j2-type"          : CUBICAL_MT[config[key].get('gain_matrix_type', 'Gain2x2')],
                    "j2-time-int"      : bsols_[0],
                    "j2-freq-int"      : bsols_[1],
                    "j2-clip-low"      : config.get('cal_gain_amplitude_clip_low', 0.5),
                    "j2-clip-high"     : config.get('cal_gain_amplitude_clip_high', 1.5),
                },  
                input=pipeline.input,
                output=pipeline.output,
                shared_memory='100Gb',
                label="{0:s}:: Calibrate step {1:d} ms={2:s}".format(step, num, msname))

    # decide which tool to use for calibration
    calwith = config.get('calibrate_with', 'meqtrees').lower()
    if calwith == 'meqtrees':
        calibrate = calibrate_meqtrees
    elif calwith == 'cubical':
        calibrate = calibrate_cubical

    # selfcal loop
    if pipeline.enable_task(config, 'image_1'):
        image(1)
    
    if pipeline.enable_task(config, 'extract_sources_1'):
        extract_sources(1)

    if pipeline.enable_task(config, 'calibrate_1'):
        calibrate(1)
        
    if pipeline.enable_task(config, 'image_2'):
        image(2)

    if pipeline.enable_task(config, 'extract_sources_2'):
        extract_sources(2)

    if pipeline.enable_task(config, 'calibrate_2'):
        calibrate(2)

    if pipeline.enable_task(config, 'image_3'):
        image(3)

    if pipeline.enable_task(config, 'extract_sources_3'):
        extract_sources(3)

    if pipeline.enable_task(config, 'calibrate_3'):
        calibrate(3)

    if pipeline.enable_task(config, 'image_4'):
        image(4)

    if pipeline.enable_task(config, 'calibrate_4'):
        calibrate(4)

    if pipeline.enable_task(config, 'image_5'):
        image(5)

    if pipeline.enable_task(config, 'restore_model'):
        num = config['restore_model']['model']

        if isinstance(num, str) and len(num.split('+'))==2:
            combine = True
            mm = num.split('+')
            models = [ '{0:s}_{1:s}-pybdsm.lsm.html:output'.format(prefix, m) for m in mm]
            final = '{0:s}_final-pybdsm.lsm.html:output'.format(prefix)

            step = 'combine_models_{0:s}_{1:s}'.format(*mm)
            recipe.add('cab/tigger_convert', step,
                {                   
                    "input-skymodel"    : models[0],
                    "append"    : models[1],
                    "output-skymodel"   : final,
                    "rename"  : True,
                    "force"   : True,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Combined models'.format(step))

        else:
            num = int(model)
            final = '{0:s}_final-pybdsm.lsm.html:output'.format(prefix)

        if config['restore_model'].get('clean_model', None):
            num = int(config['restore_model'].get('clean_model', None))
       
            conv_model = prefix + '-convolved_model.fits:output'
            recipe.add('cab/fitstool', step,
                {
                    "image"    : [prefix+'_{0:d}{2:s}-{1:s}.fits:output'.format(num, im, mfsprefix) for im in ('image','residual')],
                    "output"   : conv_model,
                    "diff"     : True,
                    "force"    : True,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Make convolved model'.format(step))

            with_cc = prefix + '-with_cc.fits:output'
            recipe.add('cab/fitstool', step,
                {
                    "image"    : [prefix+'_5{0:s}-image.fits:output'.format(mfsprefix), conv_model],
                    "output"   : with_cc,
                    "sum"      : True,
                    "force"    : True,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Add clean components'.format(step))

            recipe.add('cab/tigger_restore', step,
                {
                    "input-image"    : with_cc,
                    "input-skymodel" : final,
                    "output-image"   : prefix+'.fullrest.fits',
                    "force"          : True,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Add extracted skymodel'.format(step))

        for i,msname in enumerate(mslist):
            if pipeline.enable_task(config, 'flagging_summary'):
                step = 'flagging_summary_image_selfcal_{0:d}'.format(i)
                recipe.add('cab/casa_flagdata', step,
                    {
                      "vis"         : msname,
                      "mode"        : 'summary',
                    },
                    input=pipeline.input,
                    output=pipeline.output,
                    label='{0:s}:: Flagging summary  ms={1:s}'.format(step, msname))



NAME = 'Pre-calibration flagging'

def worker(pipeline, recipe, config):
    for i in range(pipeline.nobs):
        msname = pipeline.msnames[i]
        # flag antennas automatically based on drifts in the scan average of the 
        # auto correlation spectra per field. This doesn't strictly require any calibration. It is also
        # not field structure dependent, since it is just based on the DC of the field
        # Compares scan to median power of scans per field per channel
        # Also compares antenna to median of the array per scan per field per channel
        # This should catch any antenna with severe temperature problems
        if pipeline.enable_task(config, 'autoflag_autocorr_powerspectra'):
            step = 'autoflag_autocorr_spectra_{0:d}'.format(i)
            def_fields = ','.join([pipeline.bpcal_id[i], pipeline.gcal_id[i], pipeline.target_id[i]])
            def_calfields = ','.join([pipeline.bpcal_id[i], pipeline.gcal_id[i]])
            if config['autoflag_autocorr_powerspectra'].get('fields', 'auto') != 'auto' and \
               not set(config['autoflag_autocorr_powerspectra'].get('fields', 'auto').split(',')) <= set(['gcal', 'bpcal', 'target']):
                raise KeyError("autoflag on powerspectra fields can only be 'auto' or be a combination of 'gcal', 'bpcal' or 'target'")
            if config['autoflag_autocorr_powerspectra'].get('calibrator_fields', 'auto') != 'auto' and \
               not set(config['autoflag_autocorr_powerspectra'].get('calibrator_fields', 'auto').split(',')) <= set(['gcal', 'bpcal']):
                raise KeyError("autoflag on powerspectra calibrator fields can only be 'auto' or be a combination of 'gcal', 'bpcal'")

            fields = def_fields if config['autoflag_autocorr_powerspectra'].get('fields', 'auto') == 'auto' else \
                     ",".join([getattr(pipeline, key + "_id")[i] for key in config['autoflag_autocorr_powerspectra'].get('fields').split(',')])
            calfields = def_calfields if config['autoflag_autocorr_powerspectra'].get('calibrator_fields', 'auto') == 'auto' else \
                     ",".join([getattr(pipeline, key + "_id")[i] for key in config['autoflag_autocorr_powerspectra'].get('calibrator_fields').split(',')])

            recipe.add("cab/politsiyakat_autocorr_amp", step,
                {
                    "msname": msname,
                    "field": fields,
                    "cal_field": calfields,
                    "scan_to_scan_threshold": config["autoflag_autocorr_powerspectra"]["scan_to_scan_threshold"],
                    "antenna_to_group_threshold": config["autoflag_autocorr_powerspectra"]["antenna_to_group_threshold"],

                    "dpi": config['autoflag_autocorr_powerspectra'].get('dpi', 300),
                    "plot_size": config['autoflag_autocorr_powerspectra'].get('plot_size', 6),
                    "nproc_threads": config['autoflag_autocorr_powerspectra'].get('threads', 8),
                    "data_column": config['autoflag_autocorr_powerspectra'].get('column', "DATA")
                },
                input=pipeline.input, output=pipeline.output,
                label="{0:s}: Flag out antennas with drifts in autocorrelation powerspectra")

        if pipeline.enable_task(config, 'flag_autocorr'):
            step = 'flag_autocorr_{0:d}'.format(i)
            recipe.add('cab/casa_flagdata', step,
                {
                  "vis"         : msname,
                  "mode"        : 'manual',
                  "autocorr"    : True,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Flag auto-correlations ms={1:s}'.format(step, msname))

        if pipeline.enable_task(config, 'quack_flagging'):
            step = 'quack_flagging_{0:d}'.format(i)
            recipe.add('cab/casa_flagdata', step,
                {
                  "vis"           : msname,
                  "mode"          : 'quack',
                  "quackinterval" : config['quack_flagging'].get('quackinterval', 10),
                  "quackmode"     : config['quack_flagging'].get('quackmode', 'beg'),
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Quack flagging ms={1:s}'.format(step, msname))


        if pipeline.enable_task(config, 'flag_spw'):
            step = 'flag_spw_{0:d}'.format(i)
            recipe.add('cab/casa_flagdata','flagspw_{:d}'.format(i),
                {
                  "vis"     : msname,
                  "mode"    : 'manual',
                  "spw"     : config['flag_spw']['channels'],
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}::Flag out channels ms={1:s}'.format(step, msname))

        if pipeline.enable_task(config, 'flag_scan'):
            step = 'flag_scan_{0:d}'.format(i)
            recipe.add('cab/casa_flagdata', step,
                {
                  "vis"     : msname,
                  "mode"    : 'manual',
                  "scan"    : config['flag_scan']['scans'],
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}::Flag out channels ms={1:s}'.format(step, msname))

        if pipeline.enable_task(config, 'flag_antennas'):
            step = 'flag_antennas_{0:d}'.format(i)
            recipe.add('cab/casa_flagdata', step,
                {
                  "vis"         : msname,
                  "mode"        : 'manual',
                  "antenna"     : config['flag_antennas']['antennas']
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Flagging bad antennas ms={1:s}'.format(step, msname))

        if pipeline.enable_task(config, 'static_mask'):
            step = 'static_mask_{0:d}'.format(i)
            recipe.add('cab/rfimasker', step,
                {
                    "msname"    : msname,
                    "mask"      : config['static_mask']['mask'],
                    "accumulation_mode" : 'or',
                    "uvrange"   : config['static_mask'].get('uvrange', "''"),
                    "memory"    : 4096,
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Apply static mask ms={1:s}'.format(step, msname))

        if pipeline.enable_task(config, 'autoflag_rfi'):
            step = 'autoflag_{0:d}'.format(i)
            if config['autoflag_rfi'].get('fields', 'auto') != 'auto' and \
               not set(config['autoflag_rfi'].get('fields', 'auto').split(',')) <= set(['gcal', 'bpcal', 'target']):
                raise KeyError("autoflag rfi can only be 'auto' or be a combination of 'gcal', 'bpcal' or 'target'")
            if config['autoflag_rfi'].get('calibrator_fields', 'auto') != 'auto' and \
               not set(config['autoflag_rfi'].get('calibrator_fields', 'auto').split(',')) <= set(['gcal', 'bpcal']):
                raise KeyError("autoflag rfi fields can only be 'auto' or be a combination of 'gcal', 'bpcal'")
            def_fields = ','.join([pipeline.bpcal_id[i], pipeline.gcal_id[i], pipeline.target_id[i]])
            fields = def_fields if config['autoflag_rfi'].get('fields', 'auto') == 'auto' else \
                     ",".join([getattr(pipeline, key + "_id")[i] for key in config['autoflag_rfi'].get('fields').split(',')])

            recipe.add('cab/autoflagger', step,
                {
                  "msname"      : msname,
                  "column"      : config['autoflag_rfi'].get('column', 'DATA'),
                  # flag the calibrators for RFI and apply to target
                  "fields"      : fields,
                  "strategy"    : config['autoflag_rfi']['strategy'],
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Aoflagger flagging pass ms={1:s}'.format(step, msname))

        if pipeline.enable_task(config, 'flagging_summary'):
            step = 'flagging_summary_flagging_{0:d}'.format(i)
            recipe.add('cab/casa_flagdata', step,
                {
                  "vis"         : msname,
                  "mode"        : 'summary',
                },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Flagging summary  ms={1:s}'.format(step, msname))



#!/usr/bin/env python

import os

try:
  from setuptools import setup
except ImportError as e:
  from distutils.core import setup

requirements = [
'ruamel.yaml>=0.15.22',
'stimela>=0.3.1',
'numpy>=1.13.1',
'scipy>=0.19.1',
'pysolr>=3.4.0',
'progressbar2>=3.11.0',
'nbconvert>=5.3.1',
'aplpy>=1.1.1',
'matplotlib>=2.1.0',
'jupyter>=1.0.0',
]

PACKAGE_NAME = 'meerkathi'
__version__ = '0.1.0'

setup(name = PACKAGE_NAME,
    version = __version__,
    description = "MeerKAT end-to-end data reduction pipeline",
    author = "MeerKATHI peeps",
    author_email = "sphemakh@gmail.com",
    url = "https://github.com/sphemakh/meerkathi",
    packages=[PACKAGE_NAME],
    install_requires = requirements,
    include_package_data = True,
    scripts = ["bin/" + j for j in os.listdir("bin")],
    license=["GNU GPL v2"],
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "License :: OSI Approved :: GNU General Public License v2 (GPLv2)",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python",
        "Topic :: Scientific/Engineering :: Astronomy"
    ]
     )



import meerkathi.dispatch_crew.utils as utils
import yaml
import meerkathi
import sys

NAME = 'Automatically catergorize observed fields'

def worker(pipeline, recipe, config):

    if pipeline.enable_task(config, 'obsinfo'):
        for i in range(pipeline.nobs):
            msname = pipeline.msnames[i]
            prefix = pipeline.prefixes[i]
            if config['obsinfo'].get('listobs', True):
                step = 'listobs_{:d}'.format(i)
                recipe.add('cab/casa_listobs', step,
                    {
                      "vis"         : msname,
                      "listfile"    : prefix+'-obsinfo.txt' ,
                      "overwrite"   : True,
                    },
                    input=pipeline.input,
                    output=pipeline.output,
                    label='{0:s}:: Get observation information ms={1:s}'.format(step, msname))
        
            if config['obsinfo'].get('summary_json', True):
                 step = 'summary_json_{:d}'.format(i)
                 recipe.add('cab/msutils', step,
                    {
                      "msname"      : msname,
                      "command"     : 'summary',
                      "outfile"     : prefix+'-obsinfo.json',
                    },
                input=pipeline.input,
                output=pipeline.output,
                label='{0:s}:: Get observation information as a json file ms={1:s}'.format(step, msname))

            if config['obsinfo'].get('vampirisms', False):
                step = 'vampirisms_{0:d}'.format(i)
                recipe.add('cab/sunblocker', step,
                    {
                        "command"     : 'vampirisms',
                        "inset"       : msname,
                        "dryrun"      : True,
                        "nononsoleil" : True,
                        "verb"        : True,
                    },
                    input=pipeline.input,
                    output=pipeline.output,
                label='{0:s}:: Note sunrise and sunset'.format(step))

        recipe.run()
        recipe.jobs = []

    # itnitialse things
    for item in 'fcal bpcal gcal target reference_antenna nchans'.split():
        val = config.get(item, 'auto')
        if val and not isinstance(config[item], list):
            setattr(pipeline, item, [config[item]]*pipeline.nobs)
        elif isinstance(config[item], list):
            setattr(pipeline, item, config[item])
        else:
            setattr(pipeline, item, [None]*pipeline.nobs)

    # Set antenna properties
    pipeline.Tsys_eta = config.get('Tsys_eta', 22.0)
    pipeline.dish_diameter = config.get('dish_diameter', 13.5)
   
    for item in 'fcal bpcal gcal target'.split():
        setattr(pipeline, item + "_id", [])

    for i, prefix in enumerate(pipeline.prefixes):
        msinfo = '{0:s}/{1:s}-obsinfo.json'.format(pipeline.output, prefix)

        # get reference antenna
        if config.get('reference_antenna', 'auto') == 'auto':
            msmeta = '{0:s}/{1:s}.json'.format(pipeline.data_path, pipeline.dataid[i])
            pipeline.reference_antenna[i] = utils.meerkat_refant(msmeta)
            meerkathi.log.info('Auto selecting reference antenna as {:s}'.format(pipeline.reference_antenna[i]))

        # Get channels in MS
        if config['nchans'] == 'auto':
            with open(msinfo, 'r') as stdr:
                spw = yaml.load(stdr)['SPW']['NUM_CHAN']
                pipeline.nchans[i] = spw
            meerkathi.log.info('MS has {0:d} spectral windows, with NCHAN={1:s}'.format(len(spw), ','.join(map(str, spw))))

        #Auto select some/all fields if user didn't manually override all of them
        if 'auto' in [config[item] for item in 'fcal bpcal gcal target'.split()]:
            intents = utils.categorize_fields(msinfo)
            # Get fields and their purposes
            fcals = intents['fcal'][-1]
            gcals = intents['gcal'][-1]
            bpcals = intents['bpcal'][-1]
            targets = intents['target'][-1]

            # Set gain calibrator
            if config['gcal'] == 'auto':
                pipeline.gcal[i] = utils.select_gcal(msinfo, targets, gcals, mode='nearest')
                meerkathi.log.info('Auto selecting gain calibrator as {:s}'.format(pipeline.gcal[i]))
            else:
                pipeline.gcal[i] = config['gcal']
            tobs = utils.field_observation_length(msinfo, pipeline.gcal[i])/60.0
            meerkathi.log.info('Gain calibrator field "{0:s}" was observed for {1:.2f} minutes'.format(pipeline.gcal[i], tobs))

            # Set flux calibrator
            if config['fcal'] == 'auto':
                while len(fcals)>0:
                    fcal = utils.observed_longest(msinfo, fcals)
                    if utils.find_in_casa_calibrators(msinfo, fcal) or utils.find_in_native_calibrators(msinfo, fcal):
                        pipeline.fcal[i] = fcal
                        break
                    fcals.remove(fcal)
                meerkathi.log.info('Auto selecting flux calibrator as {:s}'.format(pipeline.fcal[i]))
            else:
                pipeline.fcal[i] = config['fcal']
            tobs = utils.field_observation_length(msinfo, pipeline.fcal[i])/60.0
            meerkathi.log.info('Flux calibrator field "{0:s}" was observed for {1:.2f} minutes'.format(pipeline.fcal[i], tobs))

            # Set bandpass calibrator
            if config['bpcal'] == 'auto':
                pipeline.bpcal[i] = utils.observed_longest(msinfo, bpcals)
                meerkathi.log.info('Auto selecting bandpass calibrator field as {:s}'.format(pipeline.bpcal[i]))
            else:
                pipeline.bpcal[i] = config['bpcal']
            tobs = utils.field_observation_length(msinfo, pipeline.bpcal[i])/60.0
            meerkathi.log.info('Bandpass calibrator field "{0:s}" was observed for {1:.2f} minutes'.format(pipeline.bpcal[i], tobs))

            # Select target field(s)
            if config['target'] == 'auto':
                pipeline.target[i] = ','.join(targets)
                meerkathi.log.info('Auto selecting target field as {:s}'.format(pipeline.target[i]))
            else:
                targets = config['target'].split(',')
                pipeline.target[i] = ','.join(targets)
            meerkathi.log.info('Found {0:d} target fields {1:s}'.format(len(targets), pipeline.target[i]))
            for target in targets:
                tobs = utils.field_observation_length(msinfo, target)/60.0
                meerkathi.log.info('Target field "{0:s}" was observed for {1:.2f} minutes'.format(target, tobs))

        # update ids for all fields now that auto fields were selected
        for item in 'fcal bpcal gcal target'.split():
            flds =  getattr(pipeline, item)[i].split(',') \
                        if isinstance(getattr(pipeline, item)[i], str) else getattr(pipeline, item)[i]
            getattr(pipeline, item + "_id").append(','.join([str(utils.get_field_id(msinfo, f)) for f in flds]))
